# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import unittest

import onnx

import onnxscript.testing
from onnxscript import ir, optimizer


def _create_model(model_text: str) -> ir.Model:
    """Create a model from the given text."""
    model = onnx.parser.parse_model(model_text)
    return ir.serde.deserialize_model(model)


class FunctionFoldingTest(unittest.TestCase):
    def test_identity(self):
        model = _create_model(
            """
            <ir_version: 7, opset_import: ["" : 17, "local" : 1]>
            agraph (float[N] x1, bool cond1) => (float[N] z1) {
                z1 = local.fun1(x1, cond1)
            }
            <opset_import: ["" : 17], domain: "local">
            fun1 (x, cond) => (z) {
                t = Identity(x)
                t2 = Identity(t)
                t3 = If (cond) <
                    then_branch = then_graph() => (t4) {
                        t5 = Identity(t2)
                        t4 = Identity(t5)
                    },
                    else_branch = else__graph() => (t6) {
                        t7 = Identity(t)
                        t6 = Identity(t7)
                    }
                >
                t4 = Add(t3, t3)
                z = Identity(t4)
            }"""
        )
        optimized = optimizer.optimize(
            model, onnx_shape_inference=False, num_iterations=1, inline=True
        )
        self.assertEqual(len(optimized.functions), 0)
        self.assertEqual(len(optimized.graph), 2)

    def test_sequence_concat(self):
        model = _create_model(
            """
            <ir_version: 7, opset_import: ["" : 17, "local" : 1]>
            agraph (float[N] x1) => (float[M] z1) {
                z1 = local.fun1(x1)
            }
            <opset_import: ["" : 17], domain: "local">
            fun1 (x) => (z) {
                t0 = Add (x, x)
                t2 = Add (x, x)
                t3 = SequenceConstruct (x, t0, t2, x)
                z = ConcatFromSequence <axis=0> (t3)
            }"""
        )
        optimized = optimizer.optimize(
            model, onnx_shape_inference=False, num_iterations=1, inline=False
        )
        function = optimized.functions[("local", "fun1", "")]
        self.assertEqual(len(function), 3)
        self.assertEqual(function[2].op_type, "Concat")

    def test_sequence_at(self):
        model = _create_model(
            """
            <ir_version: 7, opset_import: ["" : 17]>
            agraph (float[N] x) => (float[M] z) {
                t0 = Add (x, x)
                t1 = Mul (x, x)
                s = SequenceConstruct (x, t0, t1)
                one = Constant <value = int64 {1}> ()
                z = SequenceAt (s, one)
            }"""
        )
        optimized = optimizer.optimize(
            model, onnx_shape_inference=False, num_iterations=1, inline=False
        )
        expected = _create_model(
            """
            <ir_version: 7, opset_import: ["" : 17]>
            agraph (float[N] x) => (float[M] z) {
                z = Add (x, x)
            }"""
        )
        # TODO(justinchuby): Implement assert_isomorphic_graph for IR objects
        onnxscript.testing.assert_isomorphic_graph(
            ir.to_proto(optimized.graph), ir.to_proto(expected.graph)
        )

    def test_single_user_function_is_modified_inplace_after_folding(self):
        model = _create_model(
            """
            <ir_version: 7, opset_import: ["" : 17, "local" : 1]>
            agraph (float[N] x1) => (float[M] z1) {
                z1 = local.fun1(x1)
            }
            <opset_import: ["" : 17], domain: "local">
            fun1 (x) => (z) {
                t0 = Add (x, x)
                t2 = Add (x, x)
                t3 = SequenceConstruct (x, t0, t2, x)
                z = ConcatFromSequence (t3)
            }"""
        )
        optimized = optimizer.optimize(
            model, onnx_shape_inference=False, num_iterations=1, inline=False
        )
        self.assertEqual(next(iter(optimized.functions.values())).name, "fun1")

    def test_fold_nested_if_function_succeeds(self):
        model = _create_model(
            """
            <
            ir_version: 9,
            opset_import: ["this" : 1, "" : 18]
            >
            func (float[1,512] x, float[1,512] y) => ( out) {
            out = this.foldable_func (x, y)
            }
            <
            domain: "this",
            opset_import: ["" : 18]
            >
            foldable_func (x, y) => (z_6)
            {
            cond = Constant <value: tensor = bool cond {1}> ()
            z_6 = If (cond) <then_branch: graph = thenGraph_4 () => ( z_2) {
                cond_0 = Not (cond)
                z_2 = If (cond_0) <then_branch: graph = thenGraph_5 () => ( z) {
                    z = Add (x, x)
                }, else_branch: graph = elseGraph_5 () => ( z_1) {
                    z_1 = Identity (x)
                }>
            }, else_branch: graph = elseGraph_4 () => ( z_5) {
                z_5 = If (cond) <then_branch: graph = thenGraph_10 () => ( z_3) {
                    z_3 = Add (y, y)
                }, else_branch: graph = elseGraph_10 () => ( z_4) {
                    z_4 = Add (x, y)
                }>
            }>
            }"""
        )
        optimized = optimizer.optimize(model, onnx_shape_inference=False, inline=True)

        self.assertEqual(len(optimized.functions), 0)
        self.assertEqual(len(optimized.graph), 1)
        self.assertNotIn("If", {n.op_type for n in optimized.graph})


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Utilities for working with Python ASTs."""

from __future__ import annotations

import ast
import inspect
import textwrap
from typing import Callable


def get_src_and_ast(func: Callable, /) -> tuple[str, ast.FunctionDef]:
    try:
        src = inspect.getsource(func)
    except OSError as e:
        raise RuntimeError(
            f"Decorator script does not work on dynamically compiled function {func.__name__}."
        ) from e
    src = textwrap.dedent(src)
    top_level_ast = ast.parse(src)
    assert isinstance(top_level_ast, ast.Module)
    assert len(top_level_ast.body) == 1
    f_ast = top_level_ast.body[0]
    assert isinstance(f_ast, ast.FunctionDef)
    return src, f_ast


def normalize_subscript_expr(expr: ast.Subscript):
    # Normalizes the representation of a subscripted expression, handling python version
    # differences as well as variations between A[x] (single-index) and A[x, y] (multiple indices)
    # Returns a list of expressions, denoting the indices, after stripping the extraneous "Index"
    # wrapper present in python versions before 3.9
    index_expr = expr.slice
    if isinstance(index_expr, ast.Tuple):
        return index_expr.elts  # multiple indices
    else:
        return [index_expr]  # single index


def is_print_call(stmt: ast.stmt) -> bool:
    """Return True if the statement is a call to the print function."""
    if isinstance(stmt, ast.Expr):
        if isinstance(stmt.value, ast.Call):
            if isinstance(stmt.value.func, ast.Name):
                return stmt.value.func.id == "print"
    return False


def is_doc_string(stmt: ast.stmt) -> bool:
    """Return True if the statement is a docstring."""
    if isinstance(stmt, ast.Expr) and isinstance(stmt.value, ast.Constant):
        return isinstance(stmt.value.value, str)
    return False



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import pathlib

import numpy as np
import onnx
from onnx import helper as onnx_helper


def load_test_data(
    qual_model_dir: str, input_names: list[str]
) -> tuple[dict[str, np.ndarray], list[np.ndarray]]:
    test_data_dir = pathlib.Path(qual_model_dir) / "test_data_set_0"
    inputs = {}
    expected_outputs = []
    for test_data in test_data_dir.glob("input_*.pb"):
        idx = int(test_data.stem[len("input_") :])
        input_name = input_names[idx]
        input_data = onnx.TensorProto()
        with open(test_data, "rb") as f:
            input_data.ParseFromString(f.read())
        inputs[input_name] = onnx.numpy_helper.to_array(input_data)

    output_file_paths = list(test_data_dir.glob("output_*.pb"))
    expected_outputs = [None] * len(output_file_paths)
    for test_data in test_data_dir.glob("output_*.pb"):
        idx = int(test_data.stem[len("output_") :])
        output_data = onnx.TensorProto()
        with open(test_data, "rb") as f:
            output_data.ParseFromString(f.read())
        expected_outputs[idx] = onnx.numpy_helper.to_array(output_data)  # type: ignore[call-overload]

    assert all(name in inputs for name in input_names), "Some inputs are missing."
    assert not any(output is None for output in expected_outputs), "Some outputs are missing."

    return inputs, expected_outputs  # type: ignore[return-value]


def generate_random_input(model: onnx.ModelProto) -> dict[str, np.ndarray]:
    """Generate random input for the model.

    NOTE: This is unused. There is parity issue with randomly generated data. Need investigation.
    """
    inputs = {}
    for _, input in enumerate(model.graph.input):
        shape = [d.dim_value for d in input.type.tensor_type.shape.dim]
        np_dtype = onnx_helper.tensor_dtype_to_np_dtype(input.type.tensor_type.elem_type)
        if np_dtype is None:
            raise ValueError(f"Unsupported dtype: {input.type.tensor_type.elem_type}")
        if np_dtype in (np.float16, np.float32, np.float64):
            inputs[input.name] = np.random.rand(*shape).astype(np_dtype) - 0.5
        else:
            inputs[input.name] = np.random.randint(3, 100, size=shape, dtype=np_dtype)
    return inputs



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import logging

from onnxscript import ir
from onnxscript.rewriter import _ir_utils
from onnxscript.rewriter._rewrite_rule import RewriteRule, RewriteRuleSet

logger = logging.getLogger(__name__)
_INT64_MAX = 9223372036854775807


def _check_if_redundant_slice(
    context,
    data: ir.Value,
    starts: ir.Value,
    ends: ir.Value,
    axes: ir.Value,
    steps: ir.Value,
    **_,
) -> bool:
    """If the starts is 0, and the ends is equal to or grater than the shape of the specified axis, then the slice is redundant."""
    del context  # Reserved for future extensions

    starts_const = starts.const_value
    ends_const = ends.const_value
    axes_const = axes.const_value
    steps_const = steps.const_value

    if starts_const is None or ends_const is None or axes_const is None or steps_const is None:
        logger.info("The value 'start', 'end', 'axis', 'step' is not statically known.")
        return False

    # Check if the values are scalar
    if starts_const.numpy().size != 1:  # type: ignore[union-attr]
        logger.info("The value 'start' is not a scalar.")
        return False
    if ends_const.numpy().size != 1:  # type: ignore[union-attr]
        logger.info("The value 'end' is not a scalar.")
        return False
    if axes_const.numpy().size != 1:  # type: ignore[union-attr]
        logger.info("The value 'axis' is not a scalar.")
        return False
    if steps_const.numpy().size != 1:  # type: ignore[union-attr]
        logger.info("The value 'step' is not a scalar.")
        return False

    if steps_const.numpy().item() != 1:
        logger.info("The value 'step' is not 1.")
        return False
    # starts is 0
    if starts_const.numpy().item() != 0:
        logger.info("The value 'start' is not 0.")
        return False
    # In case data.shape is not statically known, we still can tell the slice is redundant if ends is sys.maxsize
    if ends_const.numpy().item() == _INT64_MAX:
        return True
    if data.shape is None or data.shape.is_dynamic(axes_const.numpy().item()):
        logger.info("The value 'data' shape is not statically known.")
        return False
    if ends_const.numpy().item() < data.shape[axes_const.numpy().item()]:
        logger.info("The value 'end' is less than the shape of the specified axis.")
        return False

    return True


def _identity_to_itself(op, data, **_):
    """Return the input data as the output."""
    return op.Identity(data)


def _potential_redundant_slice(op, data, starts, ends, axes, steps):
    """To identify a slice op"""
    return op.Slice(data, starts, ends, axes, steps, _outputs=["slice_output"])


def _same_shape(op, data: ir.Value, slice_output: ir.Value, steps: ir.Value, **_):
    """Check if the shape of the slice output is the same as the data."""
    if data.shape is None or slice_output.shape is None:
        return False

    if not _ir_utils.is_singleton_value(steps, 1):
        return False

    return _ir_utils.same_shape(data.shape, slice_output.shape)


# Register the rewrite rules
collapse_slice_rule = RewriteRule(
    _potential_redundant_slice,
    _identity_to_itself,
    _check_if_redundant_slice,
)

collapse_slice2_rule = RewriteRule(
    _potential_redundant_slice,
    _identity_to_itself,
    _same_shape,
)

# NOTE: The second rule subsumes the first one. So, we may be able to remove the first one,
# provided shape-inference is run before the rewriter and computes the shape of the slice output.

rules = RewriteRuleSet([collapse_slice_rule, collapse_slice2_rule])



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import unittest
from typing import Mapping, Sequence

import numpy as np
import onnx_ir as ir
import parameterized
from onnx_ir.passes.common import onnx_checker, shape_inference

from onnxscript.rewriter import pattern as orp
from onnxscript.rewriter import testing
from onnxscript.rewriter.rules.common import _fuse_pad_into_conv
from onnxscript.rewriter.rules.common._fuse_pad_into_conv import (
    fuse_pad_into_conv_rule,
    normalize_pad_format_conv_rule,
)


def _clone_model(model: ir.Model) -> ir.Model:
    return ir.from_proto(ir.to_proto(model))


class FuseConvPadBaseTest(unittest.TestCase):
    @property
    def rng(self):
        return np.random.default_rng(20250522)

    def get_conv_weights(self, shape: Sequence[int], tape: ir.tape.Tape = None):
        w = ir.tensor(self.rng.uniform(-0.5, 0.5, shape).astype("float32"), name="W")
        if tape is not None:
            w = tape.initializer(w)
        return w

    def build_model(
        self,
        op_type: str,
        input_shape: ir.Shape,
        weight_shape: Sequence[int],
        pad_inputs: Sequence[ir.TensorProtocol | ir.Value | None],
        pad_attributes: Mapping[str, ir.Attr] | None = None,
        conv_attributes: Mapping[str, ir.Attr] | None = None,
    ) -> ir.Model:
        tape = ir.tape.Tape()
        inputs = []
        output_shape = ir.Shape((input_shape[0],) + ("?",) * (len(input_shape) - 1))

        # Convert pad_inputs to initializers (if needed)
        pad_inputs = list(pad_inputs)
        for idx, x in enumerate(pad_inputs):
            if isinstance(x, ir.TensorProtocol):
                pad_inputs[idx] = tape.initializer(x)
            elif isinstance(x, ir.Value):
                inputs.append(x)
            elif isinstance(x, float):
                pad_inputs[idx] = tape.op("Constant", inputs=[], attributes={"value_float": x})
            elif x is not None:
                raise ValueError(f"Unsupported type for pad input ({x}): {type(x)}.")

        # Register operations in the tape
        idtype = ir.DataType.UINT8 if op_type == "ConvInteger" else ir.DataType.FLOAT
        x = ir.val("X", shape=input_shape, type=ir.TensorType(idtype))
        y = tape.op("Pad", inputs=[x, *pad_inputs], attributes=pad_attributes)
        y = tape.op(
            op_type,
            inputs=[y, self.get_conv_weights(weight_shape, tape)],
            attributes=conv_attributes,
            output=ir.val("Y", shape=output_shape, type=ir.TensorType(x.dtype)),
        )
        if op_type == "ConvInteger":
            y.dtype = ir.DataType.INT32

        # Build the model
        ir_model = ir.Model(
            ir.Graph(
                inputs=[x, *inputs],
                outputs=[y],
                nodes=tape.nodes,
                initializers=tape.initializers,
                opset_imports={"": 20},
                name="model",
            ),
            ir_version=10,
        )
        onnx_checker.CheckerPass(True)(ir_model)
        ir_model = shape_inference.infer_shapes(ir_model)
        return ir_model


class FuseConvPadTest(FuseConvPadBaseTest):
    @parameterized.parameterized.expand(
        [
            (pad_pads, const_value, axes, conv_pads, conv_auto_pad)
            for pad_pads, axes, conv_pads, conv_auto_pad in [
                ([0, 0, 2, 2, 0, 0, 2, 2], None, None, None),
                ([0, 2, 2, 0, 2, 2], ir.tensor([1, -2, -1], name="axes"), [2, 0, 2, 0], None),
                ([1, 1, 1, 1], ir.tensor([-2, 3], name="axes"), [0, 1, 0, 1], None),
                ([1, 3, 1, 3], ir.tensor([3, 2], name="axes"), None, "VALID"),
            ]
            for const_value in [None, 0.0]
        ]
    )
    def test_fuse_pad_into_conv(self, pad_pads, const_value, axes, conv_pads, conv_auto_pad):
        pad_inputs = [ir.tensor(pad_pads, name="pads")]
        if const_value is not None or axes is not None:
            pad_inputs.append(const_value)
        if axes is not None:
            pad_inputs.append(axes)
        base_model = self.build_model(
            op_type="Conv",
            input_shape=ir.Shape(("N", 32, 14, 16)),
            weight_shape=(10, 32, 3, 3),
            pad_inputs=pad_inputs,
            conv_attributes={"pads": conv_pads, "auto_pad": conv_auto_pad},
        )
        updated_model = _clone_model(base_model)

        # Apply rule
        count = _fuse_pad_into_conv.rules.apply_to_model(updated_model)

        # Check that Pad was fused
        self.assertEqual(count, 1 if conv_auto_pad is None else 2)
        self.assertEqual(updated_model.graph.num_nodes(), 1)
        onnx_checker.CheckerPass(True)(updated_model)

        # Check inference
        inputs = self.rng.random((1, 32, 14, 16), dtype="float32")
        testing.assert_numerically_equal(base_model, updated_model, (inputs,), atol=0, rtol=0)

    @parameterized.parameterized.expand(
        [
            (
                "constant",
                ir.tensor([1] * 10, name="pads"),
                ir.tensor([0.0], name="const_value"),
                None,
                "NOTSET",
                "must be zero in non-spatial dimensions",
            ),
            (
                "constant",
                ir.tensor([0, 0, 0, 0], name="pads"),
                ir.tensor([1.0], name="const_value"),
                ir.tensor([0, -1], name="axes"),
                "NOTSET",
                "must be equal to 0.",
            ),
            (
                "edge",
                ir.tensor([0, 0, 0, 0], name="pads"),
                ir.tensor([0.0], name="const_value"),
                ir.tensor([0, -1], name="axes"),
                "NOTSET",
                "mode must be 'constant'.",
            ),
            (
                "constant",
                ir.Value(
                    name="pads", shape=ir.Shape([4]), type=ir.TensorType(ir.DataType.INT64)
                ),
                None,
                ir.tensor([0, -1], name="axes"),
                "NOTSET",
                "pads is not a constant/initializer.",
            ),
            (
                "constant",
                ir.tensor([0] * 10, name="pads"),
                ir.Value(
                    name="cval", shape=ir.Shape([1]), type=ir.TensorType(ir.DataType.FLOAT)
                ),
                None,
                "NOTSET",
                "cval is not a constant",
            ),
            (
                "constant",
                ir.tensor([0, 0, 0, 0], name="pads"),
                None,
                ir.Value(
                    name="axes", shape=ir.Shape([2]), type=ir.TensorType(ir.DataType.INT64)
                ),
                "NOTSET",
                "axes is not a constant",
            ),
            (
                "constant",
                ir.tensor([0, 0, 0, 0], name="pads"),
                ir.tensor([0.0], name="const_value"),
                ir.tensor([0, -1], name="axes"),
                "VALID",
                "auto_pad must be 'NOTSET'.",
            ),
        ]
    )
    def test_unsupported_fuse_pad_into_conv(
        self, mode, pads, const_value, axes, auto_pad, err_msg
    ):
        base_model = self.build_model(
            op_type="Conv",
            input_shape=ir.Shape(("N", 32, 14, 16, 12)),
            weight_shape=(10, 32, 3, 4, 5),
            pad_inputs=[pads, const_value, axes],
            pad_attributes={"mode": mode},
            conv_attributes={"auto_pad": auto_pad},
        )

        # Apply rule and check it was not applied
        tracer = orp.MatchingTracer()
        count = fuse_pad_into_conv_rule.apply_to_model(base_model, tracer=tracer)
        self.assertEqual(count, 0)

        # Check that the error message is the expected one
        tracer_match = tracer.best_matches_map[fuse_pad_into_conv_rule][0]
        self.assertEqual(tracer_match.status.value, orp.MatchStatus.CONDITION_FAILED)
        self.assertRegex(tracer_match.match_result.reason, err_msg)


class FuseConvIntegerPadTest(FuseConvPadBaseTest):
    def get_conv_weights(self, shape: Sequence[int], tape: ir.tape.Tape = None):
        w = ir.tensor(self.rng.integers(0, 256, shape).astype("uint8"), name="W")
        if tape is not None:
            w = tape.initializer(w)
        return w

    @parameterized.parameterized.expand(
        [
            (pad_pads, const_value, axes, conv_pads, conv_auto_pad)
            for pad_pads, axes, conv_pads, conv_auto_pad in [
                ([0, 0, 3, 2, 0, 0, 1, 4], None, [1, 1, 1, 1], None),
                ([2, 2, 0, 2, 2, 0], ir.tensor([-2, -1, 1], name="axes"), None, None),
                ([1, 2, 2, 1], ir.tensor([-1, 2], name="axes"), [0, 1, 0, 1], None),
                ([3, 3], ir.tensor([2], name="axes"), None, "SAME_UPPER"),
            ]
            for const_value in [None, ir.tensor(np.array([0], "uint8"), name="const_value")]
        ]
    )
    def test_fuse_pad_into_conv_integer(
        self, pad_pads, const_value, axes, conv_pads, conv_auto_pad
    ):
        pad_inputs = [ir.tensor(pad_pads, name="pads")]
        if const_value is not None or axes is not None:
            pad_inputs.append(const_value)
        if axes is not None:
            pad_inputs.append(axes)
        base_model = self.build_model(
            op_type="ConvInteger",
            input_shape=ir.Shape(("N", 24, 19, 23)),
            weight_shape=(8, 24, 3, 3),
            pad_inputs=pad_inputs,
            conv_attributes={"pads": conv_pads, "auto_pad": conv_auto_pad},
        )
        updated_model = _clone_model(base_model)

        # Apply rule
        count = _fuse_pad_into_conv.rules.apply_to_model(updated_model)

        # Check that Pad was fused
        self.assertEqual(count, 1 if conv_auto_pad is None else 2)
        self.assertEqual(updated_model.graph.num_nodes(), 1)
        onnx_checker.CheckerPass(True)(updated_model)

        # Check inference
        inputs = self.rng.integers(0, 255, (1, 24, 19, 23), dtype="uint8")
        testing.assert_numerically_equal(base_model, updated_model, (inputs,), atol=0, rtol=0)


class NormalizePadFormatTest(FuseConvPadBaseTest):
    def build_model(
        self,
        input_shape: ir.Shape,
        conv_inputs: Sequence[int],
        conv_attributes: Mapping[str, ir.Attr] | None = None,
        infer_shapes=True,
    ) -> ir.Model:
        tape = ir.tape.Tape()
        inputs = []
        output_shape = ir.Shape(("?",) * len(input_shape))

        # Convert conv_inputs to initializers (if needed)
        conv_inputs = list(conv_inputs)
        for idx, x in enumerate(conv_inputs):
            if isinstance(x, ir.TensorProtocol):
                conv_inputs[idx] = tape.initializer(x)
            elif isinstance(x, ir.Value):
                inputs.append(x)
            elif x is not None:
                raise ValueError(f"Unsupported type for pad input ({x}): {type(x)}.")

        # Register operations in the tape
        x = ir.val("X", shape=input_shape, type=ir.TensorType(ir.DataType.FLOAT))
        y = tape.op(
            "Conv",
            inputs=[x, *conv_inputs],
            attributes=conv_attributes,
            output=ir.val("Y", shape=output_shape, type=x.type),
        )

        # Build the model
        ir_model = ir.Model(
            ir.Graph(
                inputs=[x, *inputs],
                outputs=[y],
                nodes=tape.nodes,
                initializers=tape.initializers,
                opset_imports={"": 20},
                name="model",
            ),
            ir_version=10,
        )
        if len(input_shape) > 0 and infer_shapes:
            onnx_checker.CheckerPass(True)(ir_model)
            ir_model = shape_inference.infer_shapes(ir_model)
        else:
            onnx_checker.CheckerPass(False)(ir_model)
        return ir_model

    @parameterized.parameterized.expand(
        [
            (dynamic_shape, strides, kernel_shape, auto_pad)
            for strides, kernel_shape in [((2, 3), (1, 4)), ((2, 1), (5, 2))]
            for dynamic_shape, auto_pad in [
                (False, "SAME_UPPER"),
                (False, "SAME_LOWER"),
                (True, "VALID"),
            ]
        ]
    )
    def test_normalize_pad_format(self, dynamic_shape, strides, kernel_shape, auto_pad):
        input_shape = (
            ir.Shape(("N", "A", "B", "C")) if dynamic_shape else ir.Shape(("N", 32, 22, 27))
        )
        base_model = self.build_model(
            input_shape=input_shape,
            conv_inputs=[ir.tensor(self.get_conv_weights((32, 32, *kernel_shape)), name="W")],
            conv_attributes={
                "strides": strides,
                "auto_pad": auto_pad,
                "kernel_shape": kernel_shape,
            },
        )
        updated_model = _clone_model(base_model)

        # Apply rule
        count = _fuse_pad_into_conv.rules.apply_to_model(updated_model)
        onnx_checker.CheckerPass(True)(updated_model)

        # Check conv has changed
        self.assertEqual(count, 1)
        self.assertEqual(updated_model.graph[0].attributes.get_string("auto_pad"), "NOTSET")

        # Check inference
        inputs = self.rng.random((1, 32, 22, 27), dtype="float32")
        testing.assert_numerically_equal(base_model, updated_model, (inputs,), atol=0, rtol=0)

    @parameterized.parameterized.expand(
        [
            (ir.Shape([]), False, "Input shapes are not defined"),
            (ir.Shape(("N", "C", "A")), False, "Expected static spatial input shapes"),
            (ir.Shape(("N", "C", 32)), False, "Expected static spatial output shapes"),
        ]
    )
    def test_unsupported_normalize_pad_format(self, input_shape, infer_shapes, error_msg):
        base_model = self.build_model(
            input_shape=input_shape,
            conv_inputs=[ir.tensor(np.ones((32, 11, 4)), name="W")],
            conv_attributes={"auto_pad": "SAME_UPPER"},
            infer_shapes=infer_shapes,
        )

        # Apply rule and check it was not applied
        tracer = orp.MatchingTracer()
        count = normalize_pad_format_conv_rule.apply_to_model(base_model, tracer=tracer)
        self.assertEqual(count, 0)

        # Check that the error message is the expected one
        tracer_match = tracer.best_matches_map[normalize_pad_format_conv_rule][0]
        self.assertEqual(tracer_match.status.value, orp.MatchStatus.CONDITION_FAILED)
        self.assertRegex(tracer_match.match_result.reason, error_msg)

    def test_unsupported_normalize_pad_format_on_weights(self):
        W = ir.Value(name="W", shape=ir.Shape([]), type=ir.TensorType(ir.DataType.FLOAT))
        base_model = self.build_model(
            input_shape=ir.Shape(("N", 2, 32)),
            conv_inputs=[W],
            conv_attributes={"auto_pad": "SAME_UPPER"},
            infer_shapes=False,
        )
        # Set output shape to analyze error due to weights
        base_model.graph[0].outputs[0].shape = ir.Shape(("N", 10, 32))

        # Apply rule and check it was not applied
        tracer = orp.MatchingTracer()
        count = normalize_pad_format_conv_rule.apply_to_model(base_model, tracer=tracer)
        self.assertEqual(count, 0)

        # Check that the error message is the expected one
        tracer_match = tracer.best_matches_map[normalize_pad_format_conv_rule][0]
        self.assertEqual(tracer_match.status.value, orp.MatchStatus.CONDITION_FAILED)
        self.assertRegex(tracer_match.match_result.reason, "same length than kernel_shape")


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from __future__ import annotations

from typing import Sequence

import onnx_ir as ir

from onnxscript._internal import builder as _builder


class Parameter(ir.Value):
    """A module parameter that is also an ``ir.Value``.

    Since ``Parameter`` subclasses ``ir.Value``, it can be passed directly
    to ONNX ops inside ``Module.forward()`` without any conversion.
    Calling :meth:`_realize` qualifies the name with the current module
    context and registers the parameter as a graph initializer.

    Args:
        shape: Shape of the parameter tensor.
        dtype: Data type of the parameter. If None and data is not provided, defaults to float32.
        name: Name for the parameter. If None, the attribute name from
            the parent Module is used.
        data: Optional initial tensor data. If provided, the initializer
            will carry this as its const_value.
    """

    def __init__(
        self,
        shape: Sequence[int],
        dtype: ir.DataType | None = None,
        name: str | None = None,
        data: ir.TensorProtocol | None = None,
    ) -> None:
        if data is not None:
            if dtype is not None and data.dtype != dtype:
                raise ValueError(
                    f"Data type of provided data ({data.dtype}) does not match the specified dtype ({dtype})."
                )
            if dtype is None:
                dtype = data.dtype
        elif dtype is None:
            dtype = ir.DataType.FLOAT

        super().__init__(
            name=name,
            shape=ir.Shape(shape),
            type=ir.TensorType(dtype),
            const_value=data,
        )
        self._realized = False

    @property
    def dtype(self) -> ir.DataType | None:  # type: ignore[override]
        """Return the element data type of this parameter."""
        return self.type.dtype if self.type is not None else None

    def _realize(self, builder: _builder.GraphBuilder) -> Parameter:
        """Qualify the name and register as a graph initializer.

        Uses the builder's *root* graph builder to qualify the name and
        register the initializer.  When the builder is a sub-builder (e.g.
        for a Scan body), this ensures the parameter is stored in the
        main graph — making it visible as an implicit input to the
        subgraph rather than incorrectly placed inside it.

        Uses direct assignment to ``graph.initializers[...]`` to skip the
        const_value check. Idempotent: subsequent calls are no-ops.
        """
        if self._realized:
            return self

        self_name = self.name
        if not self_name:
            raise ValueError(
                "Parameter._realize() called on a Parameter without a name. "
                "Ensure the Parameter is attached to a Module attribute or otherwise "
                "initialized with a name before realization."
            )
        root = builder.root
        self_name = self.name = root._qualify_initializer_name(self_name)  # pylint: disable=protected-access
        root.graph.initializers[self_name] = self
        self._realized = True
        return self

    def __repr__(self) -> str:
        return f"Parameter(shape={list(self.shape)}, dtype={self.dtype}, name={self.name!r})"



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import unittest

from onnxscript.rewriter import _pattern_ir


class PatternIRTest(unittest.TestCase):
    """Test _pattern_ir module functionality."""

    def test_value_pattern_with_check(self):
        """Test ValuePattern with check attribute."""

        def value_checker(context, value):
            return True

        # Test creating ValuePattern with check
        value_pattern = _pattern_ir.ValuePattern("test_value", check=value_checker)
        self.assertIs(value_pattern._check, value_checker)
        self.assertEqual(value_pattern.name, "test_value")

    def test_node_pattern_with_check(self):
        """Test NodePattern with check attribute."""

        def node_checker(context, node):
            return True

        # Test creating NodePattern with check
        domain_pattern = _pattern_ir.StringConstantPattern("")
        inputs = []
        attributes = {}
        outputs = ["output"]

        node_pattern = _pattern_ir.NodePattern(
            domain_pattern,
            "Add",
            inputs,
            attributes,
            outputs,
            allow_other_attributes=True,
            allow_other_inputs=True,
            check=node_checker,
        )
        self.assertIs(node_pattern._check, node_checker)

    def test_to_value_pattern_with_callable(self):
        """Test _to_value_pattern function with callable input."""

        def my_checker(context, value):
            return True

        result = _pattern_ir._to_value_pattern(my_checker)
        self.assertIsInstance(result, _pattern_ir.ValuePattern)
        self.assertIs(result._check, my_checker)
        self.assertIsNone(result.name)

    def test_op_pattern_builder_with_check(self):
        """Test OpPatternBuilder with _check parameter."""

        def node_checker(context, node):
            return True

        # Create OpPatternBuilder and call with _check parameter
        opset_builder = _pattern_ir.OpsetPatternBuilder("")
        result = opset_builder.Add(None, None, _check=node_checker)

        # The result should be a NodeOutputPattern, and its producer should have the check
        self.assertTrue(hasattr(result, "producer"))
        producer = result.producer()
        self.assertIsNotNone(producer)
        self.assertTrue(hasattr(producer, "_check"))
        self.assertIs(producer._check, node_checker)

    def test_graph_pattern_output_nodes_have_deterministic_order(self):
        """Test that GraphPattern.output_nodes preserves insertion order from outputs.

        Regression test for https://github.com/microsoft/onnxscript/issues/2234.
        When output_nodes was built from a set, Python's hash randomization could
        cause non-deterministic ordering, leading to non-deterministic pattern
        matching behavior for multi-output patterns.
        """
        opset_builder = _pattern_ir.OpsetPatternBuilder("")
        x = _pattern_ir.ValuePattern("x")
        # Create two distinct node patterns via two separate ops
        out_a = opset_builder.Relu(x, _outputs=["a"])
        out_b = opset_builder.Sigmoid(x, _outputs=["b"])
        outputs = [out_a, out_b]

        # Build the graph pattern multiple times and check the order is always the same
        for _ in range(50):
            graph_pattern = _pattern_ir.GraphPattern(inputs=[x], outputs=outputs, nodes=[])
            node_op_ids = [n._op_identifier for n in graph_pattern.output_nodes]
            self.assertEqual(
                node_op_ids,
                [("", "Relu", ""), ("", "Sigmoid", "")],
                "output_nodes order must match the order of outputs",
            )


if __name__ == "__main__":
    unittest.main()



# --------------------------------------------------------------------------
# ⚠️ WARNING - AUTO-GENERATED CODE - DO NOT EDIT ⚠️
# ⚙️ Generated by 'python -m opgen'
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# pylint: disable=W0221,W0222,R0901,W0237
# mypy: disable-error-code=override
# --------------------------------------------------------------------------

from __future__ import annotations

from typing import Mapping, Tuple

from onnx.defs import onnx_opset_version

from onnxscript.onnx_opset._impl.opset1 import Opset1
from onnxscript.onnx_opset._impl.opset2 import Opset2
from onnxscript.onnx_opset._impl.opset3 import Opset3
from onnxscript.onnx_opset._impl.opset4 import Opset4
from onnxscript.onnx_opset._impl.opset5 import Opset5
from onnxscript.onnx_opset._impl.opset6 import Opset6
from onnxscript.onnx_opset._impl.opset7 import Opset7
from onnxscript.onnx_opset._impl.opset8 import Opset8
from onnxscript.onnx_opset._impl.opset9 import Opset9
from onnxscript.onnx_opset._impl.opset10 import Opset10
from onnxscript.onnx_opset._impl.opset11 import Opset11
from onnxscript.onnx_opset._impl.opset12 import Opset12
from onnxscript.onnx_opset._impl.opset13 import Opset13
from onnxscript.onnx_opset._impl.opset14 import Opset14
from onnxscript.onnx_opset._impl.opset15 import Opset15
from onnxscript.onnx_opset._impl.opset16 import Opset16
from onnxscript.onnx_opset._impl.opset17 import Opset17
from onnxscript.onnx_opset._impl.opset18 import Opset18
from onnxscript.onnx_opset._impl.opset19 import Opset19
from onnxscript.onnx_opset._impl.opset20 import Opset20
from onnxscript.onnx_opset._impl.opset21 import Opset21
from onnxscript.onnx_opset._impl.opset22 import Opset22
from onnxscript.onnx_opset._impl.opset23 import Opset23
from onnxscript.onnx_opset._impl.opset24 import Opset24
from onnxscript.onnx_opset._impl.opset_ai_onnx_ml1 import Opset_ai_onnx_ml1
from onnxscript.onnx_opset._impl.opset_ai_onnx_ml2 import Opset_ai_onnx_ml2
from onnxscript.onnx_opset._impl.opset_ai_onnx_ml3 import Opset_ai_onnx_ml3
from onnxscript.onnx_opset._impl.opset_ai_onnx_ml4 import Opset_ai_onnx_ml4
from onnxscript.onnx_opset._impl.opset_ai_onnx_ml5 import Opset_ai_onnx_ml5
from onnxscript.values import Opset

__all__ = [
    "all_opsets",
    "opset1",
    "opset2",
    "opset3",
    "opset4",
    "opset5",
    "opset6",
    "opset7",
    "opset8",
    "opset9",
    "opset10",
    "opset11",
    "opset12",
    "opset13",
    "opset14",
    "opset15",
    "opset16",
    "opset17",
    "opset18",
    "opset19",
    "opset20",
    "opset21",
    "opset22",
    "opset23",
    "opset24",
    "opset_ai_onnx_ml1",
    "opset_ai_onnx_ml2",
    "opset_ai_onnx_ml3",
    "opset_ai_onnx_ml4",
    "opset_ai_onnx_ml5",
]


if onnx_opset_version() < 14:
    raise ImportError(
        f"ONNX Script requires ONNX opset >= 14 but {onnx_opset_version()} is detected."
    )


opset1 = Opset1()
opset2 = Opset2()
opset3 = Opset3()
opset4 = Opset4()
opset5 = Opset5()
opset6 = Opset6()
opset7 = Opset7()
opset8 = Opset8()
opset9 = Opset9()
opset10 = Opset10()
opset11 = Opset11()
opset12 = Opset12()
opset13 = Opset13()
opset14 = Opset14()
opset15 = Opset15()
opset16 = Opset16()
opset17 = Opset17()
opset18 = Opset18()
opset19 = Opset19()
opset20 = Opset20()
opset21 = Opset21()
opset22 = Opset22()
opset23 = Opset23()
opset24 = Opset24()
opset_ai_onnx_ml1 = Opset_ai_onnx_ml1()
opset_ai_onnx_ml2 = Opset_ai_onnx_ml2()
opset_ai_onnx_ml3 = Opset_ai_onnx_ml3()
opset_ai_onnx_ml4 = Opset_ai_onnx_ml4()
opset_ai_onnx_ml5 = Opset_ai_onnx_ml5()
all_opsets: Mapping[Tuple[str, int], Opset] = {
    (
        "",
        1,
    ): opset1,
    (
        "",
        2,
    ): opset2,
    (
        "",
        3,
    ): opset3,
    (
        "",
        4,
    ): opset4,
    (
        "",
        5,
    ): opset5,
    (
        "",
        6,
    ): opset6,
    (
        "",
        7,
    ): opset7,
    (
        "",
        8,
    ): opset8,
    (
        "",
        9,
    ): opset9,
    (
        "",
        10,
    ): opset10,
    (
        "",
        11,
    ): opset11,
    (
        "",
        12,
    ): opset12,
    (
        "",
        13,
    ): opset13,
    (
        "",
        14,
    ): opset14,
    (
        "",
        15,
    ): opset15,
    (
        "",
        16,
    ): opset16,
    (
        "",
        17,
    ): opset17,
    (
        "",
        18,
    ): opset18,
    (
        "",
        19,
    ): opset19,
    (
        "",
        20,
    ): opset20,
    (
        "",
        21,
    ): opset21,
    (
        "",
        22,
    ): opset22,
    (
        "",
        23,
    ): opset23,
    (
        "",
        24,
    ): opset24,
    (
        "ai.onnx.ml",
        1,
    ): opset_ai_onnx_ml1,
    (
        "ai.onnx.ml",
        2,
    ): opset_ai_onnx_ml2,
    (
        "ai.onnx.ml",
        3,
    ): opset_ai_onnx_ml3,
    (
        "ai.onnx.ml",
        4,
    ): opset_ai_onnx_ml4,
    (
        "ai.onnx.ml",
        5,
    ): opset_ai_onnx_ml5,
}



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Test cases for type constraint deduction functionality."""

from __future__ import annotations

import inspect
import logging
import unittest
from typing import Generator

import parameterized

import onnxscript
import onnxscript.function_libs.torch_lib.ops  # Import to populate registry  # noqa: F401
from onnxscript.function_libs.tools.torch_lib import deduce_type_constraints
from onnxscript.function_libs.torch_lib import registration

logger = logging.getLogger(__name__)


def torch_lib_onnx_functions_from_registry() -> Generator[onnxscript.OnnxFunction, None, None]:
    for op in registration.default_registry.values():
        for func in (*op.overloads, *op.complex):
            if isinstance(func, onnxscript.OnnxFunction):
                yield func


class TestDeduceTypeConstraints(unittest.TestCase):
    _SKIP_FUNCTIONS_WITH_LOOP_OR_SCAN = (
        "_aten_as_strided_onnx",
        "_aten_unfold_onnx",
        "_aten_embedding_bag_onnx",
        "_aten_embedding_bag_1d_padding_idx_onnx",
    )

    @parameterized.parameterized.expand(
        ((op,) for op in torch_lib_onnx_functions_from_registry()),
        name_func=lambda func, _, p: f"{func.__name__}_{p.args[0].name}",
    )
    def test_deduce_type_constraints_does_not_crash_for_onnx_function(
        self, onnx_function: onnxscript.OnnxFunction
    ):
        if onnx_function.name in self._SKIP_FUNCTIONS_WITH_LOOP_OR_SCAN:
            self.skipTest("Unimplemented: function contains loop or scan node.")
        try:
            signature_type_constraint = deduce_type_constraints.deduce_type_constraints(
                onnx_function
            )
        except NotImplementedError as e:
            if "Nested function" in str(e):
                self.skipTest("Unimplemented: function contains nested function.")
        else:
            logger.info(
                "Original signature: %s%s",
                onnx_function.name,
                inspect.signature(onnx_function.function),
            )
            logger.info("%s", signature_type_constraint)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Registry for aten functions."""

from __future__ import annotations

import re
import warnings
from typing import Any, Callable, Generator, Optional

import onnxscript
from onnxscript.function_libs.torch_lib import _constants

# Regex that will match "<namespace>::<op_name>[.<overload>]"
_QUALIFIED_OPERATOR_NAME_REGEX = re.compile(
    r"^(?P<namespace>[a-zA-Z0-9_]+)::(?P<name>[a-zA-Z0-9_]+)(?P<overload>\.[a-zA-Z0-9._]+)?$"
)


class OverloadedFunction:
    """Overloaded function.

    Attributes:
        name: Name of the op. E.g. "aten::add".
        overloads: Overloads function.
        complex: Support complex functions.
    """

    def __init__(self, name: str):
        self.name = name
        self.overloads: list[Any] = []
        self.complex: list[Any] = []


class Registry:
    """Registry for aten functions."""

    def __init__(self):
        self._registry: dict[str, OverloadedFunction] = {}

    def register(self, func: Any, name: str, *, complex: bool = False) -> None:
        """Register a function."""
        overloaded_function = self._registry.setdefault(name, OverloadedFunction(name))

        if complex:
            if overloaded_function.complex:
                warnings.warn(
                    f"Complex overload for '{name}' already registered: {overloaded_function.complex}.",
                    stacklevel=3,
                )
                return
            overloaded_function.complex.append(func)
        else:
            if overloaded_function.overloads:
                warnings.warn(
                    f"Real overload for '{name}' already registered: {overloaded_function.overloads}.",
                    stacklevel=3,
                )
                return
            overloaded_function.overloads.append(func)

    def __getitem__(self, name):
        return self._registry[name]

    def __contains__(self, name):
        return name in self._registry

    def __iter__(self):
        return iter(self._registry)

    def __repr__(self):
        return repr(self._registry)

    def items(self) -> Generator[tuple[str, OverloadedFunction], None, None]:
        yield from self._registry.items()

    def values(self) -> Generator[OverloadedFunction, None, None]:
        yield from self._registry.values()


# Default registry
default_registry = Registry()


def _check_and_normalize_names(name: str | tuple[str, ...]) -> tuple[str, ...]:
    names: tuple[str, ...]

    if isinstance(name, str):
        names = (name,)
    else:
        names = name
    if not isinstance(names, tuple):
        raise TypeError(f"Name must be a string or a tuple of strings, got {name}")
    for name_ in names:
        if name_.endswith(".default") or not _QUALIFIED_OPERATOR_NAME_REGEX.fullmatch(name_):
            raise ValueError(
                f"Invalid name '{name_}'. Must be in the form 'namespace::name' for default overloads "
                "or 'namespace::name.overload' for other overloads."
            )

    return names


def torch_op(
    name: str | tuple[str, ...],
    *,
    registry: Optional[Registry] = None,
    trace_only: bool = False,
    private: bool = False,
    complex: bool = False,
) -> Callable[[Callable], onnxscript.OnnxFunction | onnxscript.values.TracedOnnxFunction]:
    """Register a torch op.

    Args:
        name: Qualified ATen name of the function. E.g. "aten::relu", "aten::add.Tensor".
            Or a tuple of names e.g. ("aten::add.Scalar", "aten::add.Tensor").
            Default overloads should be specified by omitting the overload part,
            i.e. "aten::relu" instead of "aten::relu.default".
        registry: Registry to register the function to. If None, the default registry is used.
        trace_only: Whether the function should only be traced and not compiled.
        private: Whether the function is private (not directly exposed). It should
            be true for all functions with names starting with "_".
        complex: Whether the function expects complex-valued inputs.
    """
    if registry is None:
        registry = default_registry

    def wrapper(
        func: Callable,
    ) -> onnxscript.OnnxFunction | onnxscript.values.TracedOnnxFunction:
        # Compile the function
        custom_opset = onnxscript.values.Opset(domain=_constants.DOMAIN, version=1)

        processed_func: onnxscript.OnnxFunction | onnxscript.values.TracedOnnxFunction
        if trace_only:
            processed_func = onnxscript.values.TracedOnnxFunction(custom_opset, func)
        else:
            processed_func = onnxscript.script(opset=custom_opset)(func)

        assert registry is not None
        for name_ in _check_and_normalize_names(name):
            if private:
                # TODO: Remove the private tag once all functions are no longer private.
                continue
            registry.register(processed_func, name_, complex=complex)
        return processed_func

    return wrapper



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# pylint disable: protected-access
from __future__ import annotations

import ast
import inspect
import sys
from typing import Any, Callable, Optional, Sequence, TypeVar

from typing_extensions import ParamSpec

import onnxscript
from onnxscript import ir
from onnxscript._internal import ast_utils, converter, irbuilder, values

_R = TypeVar("_R")
_P = ParamSpec("_P")


def script_check(
    f: ast.FunctionDef,
    opset: values.Opset,
    global_names: dict[str, Any],
    source: str,
    default_opset: Optional[values.Opset] = None,
) -> irbuilder.IRFunction:
    """Check that a function falls into the ONNXScript subset of Python."""
    # See if conversion succeeds.
    # TODO: cleanup Converter interface/API, separating checker from
    # converter
    convert = converter.Converter(
        opset=opset,
        global_names=global_names,
        source=source,
        default_opset=default_opset,
    )
    return convert.translate_function_def(f)


def script(
    opset: Optional[values.Opset] = None,
    default_opset: Optional[values.Opset] = None,
    **kwargs: Any,
) -> Callable[[Callable[_P, _R]], onnxscript.OnnxFunction[_P, _R]]:
    """Main decorator. Declares a function as an onnx function.

    Args:
        opset: Opset the function belongs to (see :ref:`l-api-opsets`).
        default_opset: Opset to use for operators not in the function's opset.
        kwargs: Additional keyword arguments.

    Returns:
        an instance of :class:`onnxscript.values.OnnxFunction`

    Example:
    ::

        @script()
        def log2(x):
            one = op.Constant(value=make_tensor('one', TensorProto.FLOAT, [1], [1]))
            return op.Div(op.Log(x), op.CastLike(op.Log(cst), x))

    Or:

    ::

        from onnxscript.onnx_opset import opset16

        @script(opset16)
        def log2(x):
            one = op.Constant(value=make_tensor('one', TensorProto.FLOAT, [1], [1]))
            return op.Div(op.Log(x), op.CastLike(op.Log(cst), x))
    """
    opset = opset or values.Opset("this", 1)
    if not isinstance(opset, values.Opset):
        raise TypeError(
            "Script parameter must be an opset. Did you use @script instead of @script()?"
        )

    def transform(f: Callable[_P, _R]) -> onnxscript.OnnxFunction[_P, _R]:
        if not inspect.isfunction(f):
            raise TypeError("The ONNXScript decorator should be applied to functions only.")

        src, f_ast = ast_utils.get_src_and_ast(f)
        # The script should be compiled using the globals/locals at the definition site.
        # This allows the script to reference names defined outside the script,
        # which is used for a few different purposes.
        # The following is an approximate solution that works for normal use.
        module = inspect.getmodule(f)
        closure = inspect.getclosurevars(f)
        env = module.__dict__.copy()
        env.update(closure.nonlocals)
        result = script_check(f_ast, opset, env, src, default_opset=default_opset)
        # TODO: add transformations.
        return onnxscript.OnnxFunction(opset, f, result, src, kwargs)

    return transform


def graph() -> Callable[[Callable], values.OnnxClosure]:
    """A parametric decorator used to annotate nested-functions that are used
    as graph-attributes.

    Returns:
        A decorator that returns its input function, but attaches a graph_proto
        attribute representing the input function. The translation is not
        done at this time, but previously when the outer-level function
        was translated to an OnnxFunction. The decorator just looks up
        and retrieves the GraphProto representation previously generated.

    Example:
    ::

        @script()
        def cumulative_sum(X: INT64['N']):

            # Translation of cumulative_sum by @script will also translate Sum
            # into a GraphProto, which will be stored in the OnnxFunction generated
            # for cumulative_sum. At run-time (in eager-mode), the @graph decorator
            # retrieves the pre-computed GraphProto and attaches it to the Sum function.
            @graph()
            def Sum(sum_in, next):
                sum_out = sum_in + next
                scan_out = op.Identity(sum_out)
                return sum_out, scan_out
            zero = op.Constant(value_int=0)
            # The call to higher-order operator Scan below uses the above function
            # Sum as a graph-attribute.
            all_sum, result = op.Scan (zero, X, body=Sum, num_scan_inputs=1)
            return result

    """
    # This is a bit fragile. We want to get the ONNXFunction object representing
    # the outer-scope ONNXScript function from the execution stack. The caller of
    # @graph is the original script function (cumulative_sum in the above example),
    # and the caller of that function is the wrapper function/method in the
    # corresponding OnnxFunction object.
    # Currently, there is no support for eager-mode execution of nested functions,
    # so we don't need to handle doubly nested functions (e.g., a function defined
    # inside Sum in the above example).

    function_frame = sys._getframe(1)  # pylint: disable=protected-access
    wrapper_frame = sys._getframe(3)  # pylint: disable=protected-access
    onnx_function = wrapper_frame.f_locals["self"]
    nested_functions = onnx_function.function_ir.nested_functions

    def transform(f: Callable) -> values.OnnxClosure:
        return values.OnnxClosure(nested_functions[f.__name__], function_frame, f)

    return transform


def is_converted_fun(f: Any) -> bool:
    """Return True if f is a function converted by onnxscript decorator."""
    return isinstance(f, onnxscript.OnnxFunction)


def export_onnx_lib(functions: Sequence[values.OnnxFunction], filename: str) -> None:
    # Since we don't yet have LibProto defined, we use a ModelProto as a temporary
    # container for the list of functions exported as a library, with an empty graph
    # and dummy opset_imports.

    # TODO(justinchuby): This function is not well supported. We should consider removing it
    model = ir.Model(
        ir.Graph(
            inputs=[],
            outputs=[],
            nodes=[],
            opset_imports={"": 15},
        ),
        functions=[ir.serde.deserialize_function(f.to_function_proto()) for f in functions],
        ir_version=10,
        producer_name="p2o",
    )
    ir.save(model, filename)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import ast
import unittest
from typing import Any

from onnxscript._internal import analysis, ast_utils
from onnxscript._internal.sourceinfo import formatter
from onnxscript.onnx_opset import opset15 as op


class AnalysisResultsVisitor(ast.NodeVisitor):
    """Visitor class to flatten the results of liveness analysis in a pre-order traversal."""

    def __init__(self, analyzer: analysis.AstAnalyzer) -> None:
        super().__init__()
        self.results: list[Any] = []
        self.analyzer = analyzer

    def generic_visit(self, node):
        live_in = self.analyzer.live_in(node)
        if live_in is not None:
            self.results.append(live_in)
        ast.NodeVisitor.generic_visit(self, node)
        if isinstance(node, (ast.For, ast.While)):
            last = node.body[-1]
            live_out = self.analyzer.live_out(last)
            self.results.append(live_out)  # type: ignore


class TestLivenessAnalysis(unittest.TestCase):
    def analyze(self, fun):
        source, parse_tree = ast_utils.get_src_and_ast(fun)
        analyzer = analysis.AstAnalyzer(parse_tree, formatter(source))
        visitor = AnalysisResultsVisitor(analyzer)
        visitor.visit(parse_tree)
        return visitor.results

    def assertLiveness(self, fun, expected):
        self.assertEqual(self.analyze(fun), [set(x) for x in expected])

    def test_basic1(self):
        def basic_eg(x):
            # live = {x}
            y = x + 1
            # live = {y}
            x = 1
            # live = {y}
            return y + 1

        self.assertLiveness(basic_eg, [["x"], ["y"], ["y"]])

    def test_doc_string(self):
        def basic_eg(x):
            # live = {x}
            """This is a docstring."""
            # live = {x}
            y = x + 1
            # live = {y}
            x = 1
            # live = {y}
            return y + 1

        self.assertLiveness(basic_eg, [["x"], ["x"], ["y"], ["y"]])

    def test_for_loop(self):
        def loop_eg():
            # live = {}
            sum = 0.0
            # live = {sum}
            x = 0.0
            # live = {x, sum}
            for i in range(10):
                # live = {x, sum, i}
                sum = sum + i
                # live = {x, sum}
                x = x + sum * sum
                # live = {x, sum}
            # live = {x}
            return x

        self.assertLiveness(
            loop_eg,
            [
                [],
                ["sum"],
                ["x", "sum"],
                ["x", "sum", "i"],
                ["x", "sum"],
                ["x", "sum"],
                ["x"],
            ],
        )

    def test_while_loop(self):
        def while_eg(x):
            # live = {x}
            cond = x < 100
            # live = {x, cond}
            while cond:
                # live = {x}
                x = x + 2
                # live = {x}
                cond = x < 100
                # live = {x, cond}
            # live = {x}
            return x

        self.assertLiveness(
            while_eg, [["x"], ["x", "cond"], ["x"], ["x"], ["x", "cond"], ["x"]]
        )


class TestExposedUses(unittest.TestCase):
    def assertUses(self, f, expected):
        source, parse_tree = ast_utils.get_src_and_ast(f)
        analyzer = analysis.AstAnalyzer(parse_tree, formatter(source))
        result = analyzer.exposed_uses(parse_tree.body)
        self.assertEqual(result, set(expected))

    def test_basic(self):
        def f(x):
            x = x + 10
            y = 20
            z = x + y
            x = 30 + z

        self.assertUses(f, {"x"})

    def test_if(self):
        def f(x, y, z):
            if x:
                c = 10
                result = y + c
            else:
                c = 20
                result = z + c
            c30 = 30
            return result + c30

        self.assertUses(f, {"x", "y", "z"})

    def test_for_loop(self):
        def f(x, y):
            for i in range(x):
                y = y + i
            tmp = 10
            result = y + tmp
            return result

        self.assertUses(f, {"x", "y"})

    def test_while_loop(self):
        def f(x, y):
            i = 1
            while (i < 10) and (x):
                y = y + i
            tmp = y * 2
            return tmp

        self.assertUses(f, {"x", "y"})

    def test_called_function(self):
        def f(x, y):
            def nested():  # pylint: disable=unused-variable
                return y

            return op.Dummy(x, body=nested)

        self.assertUses(f, {"x", "y"})

    def test_uncalled_function(self):
        def f(x, y):
            def nested():  # pylint: disable=unused-variable
                return y

            return op.Dummy(x)

        self.assertUses(f, {"x"})

    def test_doc_string(self):
        def f(x):
            """This is a docstring."""
            x = x + 10
            y = 20
            z = x + y
            x = 30 + z

        self.assertUses(f, {"x"})


class TestAssignedVarAnalysis(unittest.TestCase):
    def assert_assigned_vars(self, f, expected: set[str]):
        source, parse_tree = ast_utils.get_src_and_ast(f)
        analyzer = analysis.AstAnalyzer(parse_tree, formatter(source))
        result = analyzer.assigned_vars(parse_tree.body)
        self.assertEqual(result, expected)

    def test_basic_defs(self):
        def f(x):
            x = x + 1
            y = x + 2
            return y

        self.assert_assigned_vars(f, {"x", "y"})

    def test_if_defs(self):
        def f(x):
            if x > 1:
                y = x + 1
                z = 2 * y
            else:
                t = x + 2
                z = 3 * t
            return z

        self.assert_assigned_vars(f, {"z", "y", "t"})

    def test_loop_defs(self):
        def f(x):
            sum = 0
            while x > 0:
                x = x - 1
                square = x * x
                sum = sum + square
            return sum

        self.assert_assigned_vars(f, {"sum", "x", "square"})

    def test_if_loop_defs(self):
        def f(x):
            if x > 0:
                sum = 0
                while x > 0:
                    x = x - 1
                    square = x * x
                    sum = sum + square
            else:
                sum = 0
            return sum

        self.assert_assigned_vars(f, {"sum", "x", "square"})

    def test_doc_string(self):
        def f(x):
            """This is a docstring."""
            x = x + 1
            y = x + 2
            return y

        self.assert_assigned_vars(f, {"x", "y"})


class ConstantIfAnalysisTest(unittest.TestCase):
    def test_constant_ifs(self):
        cond1 = True
        cond2 = False

        def f(x):
            if cond1:
                y = x + 1
            else:
                y = x + 2
            if cond2:
                z = y * 2
            else:
                z = y * 3
            if x > 0:
                w = z - 1
            else:
                w = z + 1
            return w

        source, parse_tree = ast_utils.get_src_and_ast(f)

        analyzer = analysis.AstAnalyzer(
            parse_tree, formatter(source), {"cond1": True, "cond2": False}
        )
        for node in ast.walk(parse_tree):
            if isinstance(node, ast.If):
                result = analyzer.constant_if_condition(node)
                if isinstance(node.test, ast.Name):
                    if node.test.id == "cond1":
                        self.assertEqual(result, True)
                    elif node.test.id == "cond2":
                        self.assertEqual(result, False)
                else:
                    self.assertIsNone(result)


if __name__ == "__main__":
    unittest.main(verbosity=2)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import unittest

import numpy as np
import onnx
import onnx_ir as ir
import onnxruntime as ort
from onnx_ir.passes.common import onnx_checker, shape_inference

from onnxscript import optimizer
from onnxscript.rewriter import testing
from onnxscript.rewriter.rules.common import fuse_hardswish_rules


class FuseHardSwishTest(unittest.TestCase):
    @property
    def rng(self):
        return np.random.default_rng(20250621)

    def clone_model(self, model: ir.Model) -> ir.Model:
        return ir.from_proto(ir.to_proto(model))

    def run_test(
        self,
        base_model: ir.Model,
        expected_op_types: list[str],
        dtype: str = "float",
    ):
        onnx_checker.CheckerPass(True)(base_model)
        base_model = shape_inference.infer_shapes(base_model)
        updated_model = self.clone_model(base_model)
        _ = fuse_hardswish_rules().apply_to_model(updated_model)

        # Polish model to remove unused constants
        updated_model = optimizer.optimize(updated_model)

        # Check expected op_types
        self.assertEqual([node.op_type for node in updated_model.graph], expected_op_types)

        # Check inference
        inputs = (self.rng.integers(low=-10, high=10, size=(2 * 32), dtype=np.int32),)
        if dtype == "float":
            inputs = (inputs[0].astype(np.float32),)

        testing.assert_numerically_equal(
            base_model,
            updated_model,
            inputs,
            ort_optimization_level=ort.GraphOptimizationLevel.ORT_DISABLE_ALL,
        )

        # Validate serialized model
        output_model_proto = ir.to_proto(updated_model)
        onnx.checker.check_model(output_model_proto, full_check=True)

    def test_hardsigmoid_fusion(self):
        model_text = """
            <ir_version: 8, opset_import: ["" : 18]>
            hardsigmoid (float[N] x) => (float[N] y) {
                three = Constant <value = float[1] {3.0}> ()
                six = Constant <value = float[1] {6.0}> ()
                zero = Constant <value = float[1] {0.0}> ()
                x_plus_3 = Add(x, three)
                clipped = Clip(x_plus_3, zero, six)
                y = Div(clipped, six)
            }
        """
        model = ir.from_onnx_text(model_text)
        self.run_test(model, ["HardSigmoid"])

    def test_hardswish_fusion(self):
        model_text = """
            <ir_version: 8, opset_import: ["" : 18]>
            hardswish (float[N] x) => (float[N] y) {
                three = Constant <value = float[1] {3.0}> ()
                six = Constant <value = float[1] {6.0}> ()
                zero = Constant <value = float[1] {0.0}> ()
                x_plus_3 = Add(x, three)
                clipped = Clip(x_plus_3, zero, six)
                mul_x = Mul(clipped, x)
                y = Div(mul_x, six)
            }
        """
        model = ir.from_onnx_text(model_text)
        self.run_test(model, ["HardSwish"])

    def test_hardswish_fusion_mul_last(self):
        model_text = """
            <ir_version: 8, opset_import: ["" : 18]>
            hardswish (float[N] x) => (float[N] y) {
                three = Constant <value = float[1] {3.0}> ()
                six = Constant <value = float[1] {6.0}> ()
                zero = Constant <value = float[1] {0.0}> ()
                x_plus_3 = Add(x, three)
                clipped = Clip(x_plus_3, zero, six)
                div_x = Div(clipped, six)
                y = Mul(div_x, x)
            }
        """
        model = ir.from_onnx_text(model_text)
        self.run_test(model, ["HardSwish"])

    def test_hardswish_fusion_from_sigmoid(self):
        model_text = """
            <ir_version: 8, opset_import: ["" : 18]>
            hardswish (float[N] x) => (float[N] y) {
                hardsigmoid_out = HardSigmoid<alpha=0.16666666, beta=0.5>(x)
                y = Mul(hardsigmoid_out, x)
            }
        """
        model = ir.from_onnx_text(model_text)
        self.run_test(model, ["HardSwish"])


if __name__ == "__main__":
    unittest.main()



# --------------------------------------------------------------------------
# ⚠️ WARNING - AUTO-GENERATED CODE - DO NOT EDIT ⚠️
# ⚙️ Generated by 'python -m opgen'
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# pylint: disable=W0221,W0222,R0901,W0237
# mypy: disable-error-code=override
# ruff: noqa: D402, D405
# --------------------------------------------------------------------------

from __future__ import annotations

from typing import Optional, Sequence, Tuple, TypeVar

from onnx.defs import get_schema
from typing_extensions import TypeAlias

from onnxscript.onnx_opset._impl.opset13 import Opset13
from onnxscript.onnx_types import (
    BFLOAT16,
    BOOL,
    COMPLEX64,
    COMPLEX128,
    DOUBLE,
    FLOAT,
    FLOAT16,
    INT8,
    INT16,
    INT32,
    INT64,
    STRING,
    UINT8,
    UINT16,
    UINT32,
    UINT64,
)
from onnxscript.values import Op, Opset


class Opset14(Opset13):
    def __new__(cls):
        return Opset.__new__(cls, "", 14)

    T_Add = TypeVar(
        "T_Add",
        BFLOAT16,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Add(self, A: T_Add, B: T_Add) -> T_Add:
        r"""[🌐 Add(14)](https://onnx.ai/onnx/operators/onnx__Add.html#add-14 "Online Documentation")


        Performs element-wise binary addition (with Numpy-style broadcasting support).

        This operator supports **multidirectional (i.e., Numpy-style) broadcasting**; for more details please check `Broadcasting in ONNX <https://github.com/onnx/onnx/blob/master/docs/Broadcasting.md>`_.

        (Opset 14 change): Extend supported types to include uint8, int8, uint16, and int16.


        Args:
            A: (differentiable) First operand.

            B: (differentiable) Second operand.
        """

        schema = get_schema("Add", 14, "")
        op = Op(self, "Add", schema)
        return op(*self._prepare_inputs(schema, A, B))

    T_BatchNormalization = TypeVar("T_BatchNormalization", BFLOAT16, DOUBLE, FLOAT, FLOAT16)

    U_BatchNormalization = TypeVar("U_BatchNormalization", BFLOAT16, DOUBLE, FLOAT, FLOAT16)

    def BatchNormalization(
        self,
        X: T_BatchNormalization,
        scale: T_BatchNormalization,
        B: T_BatchNormalization,
        input_mean: U_BatchNormalization,
        input_var: U_BatchNormalization,
        *,
        epsilon: float = 9.999999747378752e-06,
        momentum: float = 0.8999999761581421,
        training_mode: int = 0,
    ) -> Tuple[T_BatchNormalization, U_BatchNormalization, U_BatchNormalization]:
        r"""[🌐 BatchNormalization(14)](https://onnx.ai/onnx/operators/onnx__BatchNormalization.html#batchnormalization-14 "Online Documentation")


        Carries out batch normalization as described in the paper
        https://arxiv.org/abs/1502.03167. Depending on the mode it is being run,
        There are five required inputs 'X', 'scale', 'B', 'input_mean' and
        'input_var'.
        Note that 'input_mean' and 'input_var' are expected to be the estimated
        statistics in inference mode (training_mode=False, default),
        and the running statistics in training mode (training_mode=True).
        There are multiple cases for the number of outputs, which we list below:

        Output case #1: Y, running_mean, running_var (training_mode=True)
        Output case #2: Y (training_mode=False)

        When training_mode=False, extra outputs are invalid.
        The outputs are updated as follows when training_mode=True:
        ::

            running_mean = input_mean * momentum + current_mean * (1 - momentum)
            running_var = input_var * momentum + current_var * (1 - momentum)

            Y = (X - current_mean) / sqrt(current_var + epsilon) * scale + B

            where:

            current_mean = ReduceMean(X, axis=all_except_channel_index)
            current_var =  ReduceVar(X, axis=all_except_channel_index)

            Notice that ReduceVar refers to the population variance, and it equals to
            sum(sqrd(x_i - x_avg)) / N
            where N is the population size (this formula does not use sample size N - 1).




        When training_mode=False:
        ::

            Y = (X - input_mean) / sqrt(input_var + epsilon) * scale + B



        For previous (depreciated) non-spatial cases, implementors are suggested
        to flatten the input shape to (N x C * D1 * D2 * ... * Dn) before a BatchNormalization Op.
        This operator has **optional** inputs/outputs. See `ONNX <https://github.com/onnx/onnx/blob/master/docs/IR.md>`_ for more details about the representation of optional arguments. An empty string may be used in the place of an actual argument's name to indicate a missing argument. Trailing optional arguments (those not followed by an argument that is present) may also be simply omitted.


        Args:
            X: (differentiable) Input data tensor from the previous operator; dimensions
                are in the form of (N x C x D1 x D2 ... Dn), where N is the batch size,
                C is the number of channels. Statistics are computed for every channel
                of C over N and D1 to Dn dimensions. For image data, input dimensions
                become (N x C x H x W). The op also accepts single dimension input of
                size N in which case C is assumed to be 1

            scale: (differentiable) Scale tensor of shape (C).

            B: (differentiable) Bias tensor of shape (C).

            input_mean: (differentiable) running (training) or estimated (testing) mean
                tensor of shape (C).

            input_var: (differentiable) running (training) or estimated (testing)
                variance tensor of shape (C).

            epsilon: The epsilon value to use to avoid division by zero.

            momentum: Factor used in computing the running mean and variance.e.g.,
                running_mean = running_mean * momentum + mean * (1 - momentum).

            training_mode: If set to true, it indicates BatchNormalization is being used
                for training, and outputs 1, 2, 3, and 4 would be populated.
        """

        schema = get_schema("BatchNormalization", 14, "")
        op = Op(self, "BatchNormalization", schema)
        return op(
            *self._prepare_inputs(schema, X, scale, B, input_mean, input_var),
            epsilon=epsilon,
            momentum=momentum,
            training_mode=training_mode,
        )

    T_CumSum = TypeVar(
        "T_CumSum", BFLOAT16, DOUBLE, FLOAT, FLOAT16, INT32, INT64, UINT32, UINT64
    )

    T2_CumSum = TypeVar("T2_CumSum", INT32, INT64)

    def CumSum(
        self, x: T_CumSum, axis: T2_CumSum, *, exclusive: int = 0, reverse: int = 0
    ) -> T_CumSum:
        r"""[🌐 CumSum(14)](https://onnx.ai/onnx/operators/onnx__CumSum.html#cumsum-14 "Online Documentation")


        Performs cumulative sum of the input elements along the given axis.
        By default, it will do the sum inclusively meaning the first element is copied as is.
        Through an `exclusive` attribute, this behavior can change to exclude the first element.
        It can also perform summation in the opposite direction of the axis. For that, set `reverse` attribute to 1.

        Example:
        ::

            input_x = [1, 2, 3]
            axis=0
            output = [1, 3, 6]
            exclusive=1
            output = [0, 1, 3]
            exclusive=0
            reverse=1
            output = [6, 5, 3]
            exclusive=1
            reverse=1
            output = [5, 3, 0]




        Args:
            x: (differentiable) An input tensor that is to be processed.

            axis: (non-differentiable) A 0-D tensor. Must be in the range [-rank(x),
                rank(x)-1]. Negative value means counting dimensions from the back.

            exclusive: If set to 1 will return exclusive sum in which the top element is
                not included. In other terms, if set to 1, the j-th output element would
                be the sum of the first (j-1) elements. Otherwise, it would be the sum
                of the first j elements.

            reverse: If set to 1 will perform the sums in reverse direction.
        """

        schema = get_schema("CumSum", 14, "")
        op = Op(self, "CumSum", schema)
        return op(*self._prepare_inputs(schema, x, axis), exclusive=exclusive, reverse=reverse)

    T_Div = TypeVar(
        "T_Div",
        BFLOAT16,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Div(self, A: T_Div, B: T_Div) -> T_Div:
        r"""[🌐 Div(14)](https://onnx.ai/onnx/operators/onnx__Div.html#div-14 "Online Documentation")


        Performs element-wise binary division (with Numpy-style broadcasting support).

        This operator supports **multidirectional (i.e., Numpy-style) broadcasting**; for more details please check `Broadcasting in ONNX <https://github.com/onnx/onnx/blob/master/docs/Broadcasting.md>`_.

        (Opset 14 change): Extend supported types to include uint8, int8, uint16, and int16.


        Args:
            A: (differentiable) First operand.

            B: (differentiable) Second operand.
        """

        schema = get_schema("Div", 14, "")
        op = Op(self, "Div", schema)
        return op(*self._prepare_inputs(schema, A, B))

    T_GRU = TypeVar("T_GRU", DOUBLE, FLOAT, FLOAT16)

    T1_GRU: TypeAlias = INT32

    def GRU(
        self,
        X: T_GRU,
        W: T_GRU,
        R: T_GRU,
        B: Optional[T_GRU] = None,
        sequence_lens: Optional[T1_GRU] = None,
        initial_h: Optional[T_GRU] = None,
        *,
        activation_alpha: Optional[Sequence[float]] = None,
        activation_beta: Optional[Sequence[float]] = None,
        activations: Optional[Sequence[str]] = None,
        clip: Optional[float] = None,
        direction: str = "forward",
        hidden_size: Optional[int] = None,
        layout: int = 0,
        linear_before_reset: int = 0,
    ) -> Tuple[T_GRU, T_GRU]:
        r"""[🌐 GRU(14)](https://onnx.ai/onnx/operators/onnx__GRU.html#gru-14 "Online Documentation")


        Computes an one-layer GRU. This operator is usually supported via some custom
        implementation such as CuDNN.

        Notations:

        * `X` - input tensor
        * `z` - update gate
        * `r` - reset gate
        * `h` - hidden gate
        * `t` - time step (t-1 means previous time step)
        * `W[zrh]` - W parameter weight matrix for update, reset, and hidden gates
        * `R[zrh]` - R recurrence weight matrix for update, reset, and hidden gates
        * `Wb[zrh]` - W bias vectors for update, reset, and hidden gates
        * `Rb[zrh]` - R bias vectors for update, reset, and hidden gates
        * `WB[zrh]` - W parameter weight matrix for backward update, reset, and hidden gates
        * `RB[zrh]` - R recurrence weight matrix for backward update, reset, and hidden gates
        * `WBb[zrh]` - W bias vectors for backward update, reset, and hidden gates
        * `RBb[zrh]` - R bias vectors for backward update, reset, and hidden gates
        * `H` - Hidden state
        * `num_directions` - 2 if direction == bidirectional else 1

        Activation functions:

        * Relu(x)                - max(0, x)
        * Tanh(x)                - (1 - e^{-2x})/(1 + e^{-2x})
        * Sigmoid(x)             - 1/(1 + e^{-x})

        NOTE:
          Below are optional

        * Affine(x)              - alpha * x + beta
        * LeakyRelu(x)           - x if x >= 0 else alpha * x
        * ThresholdedRelu(x)     - x if x >= alpha else 0
        * ScaledTanh(x)          - alpha * Tanh(beta * x)
        * HardSigmoid(x)         - min(max(alpha * x + beta, 0), 1)
        * Elu(x)                 - x if x >= 0 else alpha * (e^x - 1)
        * Softsign(x)            - x/(1 + |x|)
        * Softplus(x)            - log(1 + e^x)

        Equations (Default: f=Sigmoid, g=Tanh):

        * zt = f(Xt*(Wz^T) + Ht-1*(Rz^T) + Wbz + Rbz)
        * rt = f(Xt*(Wr^T) + Ht-1*(Rr^T) + Wbr + Rbr)
        * ht = g(Xt*(Wh^T) + (rt (.) Ht-1)*(Rh^T) + Rbh + Wbh) # default, when linear_before_reset = 0
        * ht = g(Xt*(Wh^T) + (rt (.) (Ht-1*(Rh^T) + Rbh)) + Wbh) # when linear_before_reset != 0
        * Ht = (1 - zt) (.) ht + zt (.) Ht-1
        This operator has **optional** inputs/outputs. See `ONNX <https://github.com/onnx/onnx/blob/master/docs/IR.md>`_ for more details about the representation of optional arguments. An empty string may be used in the place of an actual argument's name to indicate a missing argument. Trailing optional arguments (those not followed by an argument that is present) may also be simply omitted.


        Args:
            X: (differentiable) The input sequences packed (and potentially padded) into
                one 3-D tensor with the shape of `[seq_length, batch_size, input_size]`.

            W: (differentiable) The weight tensor for the gates. Concatenation of
                `W[zrh]` and `WB[zrh]` (if bidirectional) along dimension 0. This tensor
                has shape `[num_directions, 3*hidden_size, input_size]`.

            R: (differentiable) The recurrence weight tensor. Concatenation of `R[zrh]`
                and `RB[zrh]` (if bidirectional) along dimension 0. This tensor has
                shape `[num_directions, 3*hidden_size, hidden_size]`.

            B: (optional, differentiable) The bias tensor for the gates. Concatenation
                of `[Wb[zrh], Rb[zrh]]` and `[WBb[zrh], RBb[zrh]]` (if bidirectional)
                along dimension 0. This tensor has shape `[num_directions,
                6*hidden_size]`. Optional: If not specified - assumed to be 0

            sequence_lens: (optional, non-differentiable) Optional tensor specifying
                lengths of the sequences in a batch. If not specified - assumed all
                sequences in the batch to have length `seq_length`. It has shape
                `[batch_size]`.

            initial_h: (optional, non-differentiable) Optional initial value of the
                hidden. If not specified - assumed to be 0. It has shape
                `[num_directions, batch_size, hidden_size]`.

            activation_alpha: Optional scaling values used by some activation functions.
                The values are consumed in the order of activation functions, for
                example (f, g, h) in LSTM. Default values are the same as of
                corresponding ONNX operators.For example with LeakyRelu, the default
                alpha is 0.01.

            activation_beta: Optional scaling values used by some activation functions.
                The values are consumed in the order of activation functions, for
                example (f, g, h) in LSTM. Default values are the same as of
                corresponding ONNX operators.

            activations: A list of 2 (or 4 if bidirectional) activation functions for
                update, reset, and hidden gates. The activation functions must be one of
                the activation functions specified above. Optional: See the equations
                for default if not specified.

            clip: Cell clip threshold. Clipping bounds the elements of a tensor in the
                range of [-threshold, +threshold] and is applied to the input of
                activations. No clip if not specified.

            direction: Specify if the RNN is forward, reverse, or bidirectional. Must be
                one of forward (default), reverse, or bidirectional.

            hidden_size: Number of neurons in the hidden layer

            layout: The shape format of inputs X, initial_h and outputs Y, Y_h. If 0,
                the following shapes are expected: X.shape = [seq_length, batch_size,
                input_size], Y.shape = [seq_length, num_directions, batch_size,
                hidden_size], initial_h.shape = Y_h.shape = [num_directions, batch_size,
                hidden_size]. If 1, the following shapes are expected: X.shape =
                [batch_size, seq_length, input_size], Y.shape = [batch_size, seq_length,
                num_directions, hidden_size], initial_h.shape = Y_h.shape = [batch_size,
                num_directions, hidden_size].

            linear_before_reset: When computing the output of the hidden gate, apply the
                linear transformation before multiplying by the output of the reset
                gate.
        """

        schema = get_schema("GRU", 14, "")
        op = Op(self, "GRU", schema)
        return op(
            *self._prepare_inputs(schema, X, W, R, B, sequence_lens, initial_h),
            activation_alpha=activation_alpha,
            activation_beta=activation_beta,
            activations=activations,
            clip=clip,
            direction=direction,
            hidden_size=hidden_size,
            layout=layout,
            linear_before_reset=linear_before_reset,
        )

    T_HardSwish = TypeVar("T_HardSwish", DOUBLE, FLOAT, FLOAT16)

    def HardSwish(self, X: T_HardSwish) -> T_HardSwish:
        r"""[🌐 HardSwish(14)](https://onnx.ai/onnx/operators/onnx__HardSwish.html#hardswish-14 "Online Documentation")


        HardSwish takes one input data (Tensor<T>) and produces one output data (Tensor<T>) where
        the HardSwish function, y = x * max(0, min(1, alpha * x + beta)) = x * HardSigmoid<alpha, beta>(x),
        where alpha = 1/6 and beta = 0.5, is applied to the tensor elementwise.


        Args:
            X: (differentiable) Input tensor
        """

        schema = get_schema("HardSwish", 14, "")
        op = Op(self, "HardSwish", schema)
        return op(*self._prepare_inputs(schema, X))

    V_Identity = TypeVar(
        "V_Identity",
        Sequence[BOOL],
        Sequence[COMPLEX128],
        Sequence[COMPLEX64],
        Sequence[DOUBLE],
        Sequence[FLOAT],
        Sequence[FLOAT16],
        Sequence[INT16],
        Sequence[INT32],
        Sequence[INT64],
        Sequence[INT8],
        Sequence[STRING],
        Sequence[UINT16],
        Sequence[UINT32],
        Sequence[UINT64],
        Sequence[UINT8],
        BFLOAT16,
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Identity(self, input: V_Identity) -> V_Identity:
        r"""[🌐 Identity(14)](https://onnx.ai/onnx/operators/onnx__Identity.html#identity-14 "Online Documentation")

        Identity operator

        Args:
            input: (differentiable) Input tensor
        """

        schema = get_schema("Identity", 14, "")
        op = Op(self, "Identity", schema)
        return op(*self._prepare_inputs(schema, input))

    T_LSTM = TypeVar("T_LSTM", DOUBLE, FLOAT, FLOAT16)

    T1_LSTM: TypeAlias = INT32

    def LSTM(
        self,
        X: T_LSTM,
        W: T_LSTM,
        R: T_LSTM,
        B: Optional[T_LSTM] = None,
        sequence_lens: Optional[T1_LSTM] = None,
        initial_h: Optional[T_LSTM] = None,
        initial_c: Optional[T_LSTM] = None,
        P: Optional[T_LSTM] = None,
        *,
        activation_alpha: Optional[Sequence[float]] = None,
        activation_beta: Optional[Sequence[float]] = None,
        activations: Optional[Sequence[str]] = None,
        clip: Optional[float] = None,
        direction: str = "forward",
        hidden_size: Optional[int] = None,
        input_forget: int = 0,
        layout: int = 0,
    ) -> Tuple[T_LSTM, T_LSTM, T_LSTM]:
        r"""[🌐 LSTM(14)](https://onnx.ai/onnx/operators/onnx__LSTM.html#lstm-14 "Online Documentation")


        Computes an one-layer LSTM. This operator is usually supported via some
        custom implementation such as CuDNN.

        Notations:

        * `X` - input tensor
        * `i` - input gate
        * `o` - output gate
        * `f` - forget gate
        * `c` - cell gate
        * `t` - time step (t-1 means previous time step)
        * `W[iofc]` - W parameter weight matrix for input, output, forget, and cell gates
        * `R[iofc]` - R recurrence weight matrix for input, output, forget, and cell gates
        * `Wb[iofc]` - W bias vectors for input, output, forget, and cell gates
        * `Rb[iofc]` - R bias vectors for input, output, forget, and cell gates
        * `P[iof]`  - P peephole weight vector for input, output, and forget gates
        * `WB[iofc]` - W parameter weight matrix for backward input, output, forget, and cell gates
        * `RB[iofc]` - R recurrence weight matrix for backward input, output, forget, and cell gates
        * `WBb[iofc]` - W bias vectors for backward input, output, forget, and cell gates
        * `RBb[iofc]` - R bias vectors for backward input, output, forget, and cell gates
        * `PB[iof]`  - P peephole weight vector for backward input, output, and forget gates
        * `H` - Hidden state
        * `num_directions` - 2 if direction == bidirectional else 1

        Activation functions:

        * Relu(x)                - max(0, x)
        * Tanh(x)                - (1 - e^{-2x})/(1 + e^{-2x})
        * Sigmoid(x)             - 1/(1 + e^{-x})

        NOTE: Below are optional

        * Affine(x)              - alpha*x + beta
        * LeakyRelu(x)           - x if x >= 0 else alpha * x
        * ThresholdedRelu(x)     - x if x >= alpha else 0
        * ScaledTanh(x)          - alpha*Tanh(beta*x)
        * HardSigmoid(x)         - min(max(alpha*x + beta, 0), 1)
        * Elu(x)                 - x if x >= 0 else alpha*(e^x - 1)
        * Softsign(x)            - x/(1 + |x|)
        * Softplus(x)            - log(1 + e^x)

        Equations (Default: f=Sigmoid, g=Tanh, h=Tanh):

        * it = f(Xt*(Wi^T) + Ht-1*(Ri^T) + Pi (.) Ct-1 + Wbi + Rbi)
        * ft = f(Xt*(Wf^T) + Ht-1*(Rf^T) + Pf (.) Ct-1 + Wbf + Rbf)
        * ct = g(Xt*(Wc^T) + Ht-1*(Rc^T) + Wbc + Rbc)
        * Ct = ft (.) Ct-1 + it (.) ct
        * ot = f(Xt*(Wo^T) + Ht-1*(Ro^T) + Po (.) Ct + Wbo + Rbo)
        * Ht = ot (.) h(Ct)
        This operator has **optional** inputs/outputs. See `ONNX <https://github.com/onnx/onnx/blob/master/docs/IR.md>`_ for more details about the representation of optional arguments. An empty string may be used in the place of an actual argument's name to indicate a missing argument. Trailing optional arguments (those not followed by an argument that is present) may also be simply omitted.


        Args:
            X: (differentiable) The input sequences packed (and potentially padded) into
                one 3-D tensor with the shape of `[seq_length, batch_size, input_size]`.

            W: (differentiable) The weight tensor for the gates. Concatenation of
                `W[iofc]` and `WB[iofc]` (if bidirectional) along dimension 0. The
                tensor has shape `[num_directions, 4*hidden_size, input_size]`.

            R: (differentiable) The recurrence weight tensor. Concatenation of `R[iofc]`
                and `RB[iofc]` (if bidirectional) along dimension 0. This tensor has
                shape `[num_directions, 4*hidden_size, hidden_size]`.

            B: (optional, differentiable) The bias tensor for input gate. Concatenation
                of `[Wb[iofc], Rb[iofc]]`, and `[WBb[iofc], RBb[iofc]]` (if
                bidirectional) along dimension 0. This tensor has shape
                `[num_directions, 8*hidden_size]`. Optional: If not specified - assumed
                to be 0.

            sequence_lens: (optional, non-differentiable) Optional tensor specifying
                lengths of the sequences in a batch. If not specified - assumed all
                sequences in the batch to have length `seq_length`. It has shape
                `[batch_size]`.

            initial_h: (optional, non-differentiable) Optional initial value of the
                hidden. If not specified - assumed to be 0. It has shape
                `[num_directions, batch_size, hidden_size]`.

            initial_c: (optional, non-differentiable) Optional initial value of the
                cell. If not specified - assumed to be 0. It has shape `[num_directions,
                batch_size, hidden_size]`.

            P: (optional, differentiable) The weight tensor for peepholes. Concatenation
                of `P[iof]` and `PB[iof]` (if bidirectional) along dimension 0. It has
                shape `[num_directions, 3*hidde_size]`. Optional: If not specified -
                assumed to be 0.

            activation_alpha: Optional scaling values used by some activation functions.
                The values are consumed in the order of activation functions, for
                example (f, g, h) in LSTM. Default values are the same as of
                corresponding ONNX operators.For example with LeakyRelu, the default
                alpha is 0.01.

            activation_beta: Optional scaling values used by some activation functions.
                The values are consumed in the order of activation functions, for
                example (f, g, h) in LSTM. Default values are the same as of
                corresponding ONNX operators.

            activations: A list of 3 (or 6 if bidirectional) activation functions for
                input, output, forget, cell, and hidden. The activation functions must
                be one of the activation functions specified above. Optional: See the
                equations for default if not specified.

            clip: Cell clip threshold. Clipping bounds the elements of a tensor in the
                range of [-threshold, +threshold] and is applied to the input of
                activations. No clip if not specified.

            direction: Specify if the RNN is forward, reverse, or bidirectional. Must be
                one of forward (default), reverse, or bidirectional.

            hidden_size: Number of neurons in the hidden layer

            input_forget: Couple the input and forget gates if 1.

            layout: The shape format of inputs X, initial_h, initial_c and outputs Y,
                Y_h, Y_c. If 0, the following shapes are expected: X.shape =
                [seq_length, batch_size, input_size], Y.shape = [seq_length,
                num_directions, batch_size, hidden_size], initial_h.shape = Y_h.shape =
                initial_c.shape = Y_c.shape = [num_directions, batch_size, hidden_size].
                If 1, the following shapes are expected: X.shape = [batch_size,
                seq_length, input_size], Y.shape = [batch_size, seq_length,
                num_directions, hidden_size], initial_h.shape = Y_h.shape =
                initial_c.shape = Y_c.shape = [batch_size, num_directions, hidden_size].
        """

        schema = get_schema("LSTM", 14, "")
        op = Op(self, "LSTM", schema)
        return op(
            *self._prepare_inputs(schema, X, W, R, B, sequence_lens, initial_h, initial_c, P),
            activation_alpha=activation_alpha,
            activation_beta=activation_beta,
            activations=activations,
            clip=clip,
            direction=direction,
            hidden_size=hidden_size,
            input_forget=input_forget,
            layout=layout,
        )

    T_Mul = TypeVar(
        "T_Mul",
        BFLOAT16,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Mul(self, A: T_Mul, B: T_Mul) -> T_Mul:
        r"""[🌐 Mul(14)](https://onnx.ai/onnx/operators/onnx__Mul.html#mul-14 "Online Documentation")


        Performs element-wise binary multiplication (with Numpy-style broadcasting support).

        This operator supports **multidirectional (i.e., Numpy-style) broadcasting**; for more details please check `Broadcasting in ONNX <https://github.com/onnx/onnx/blob/master/docs/Broadcasting.md>`_.

        (Opset 14 change): Extend supported types to include uint8, int8, uint16, and int16.


        Args:
            A: (differentiable) First operand.

            B: (differentiable) Second operand.
        """

        schema = get_schema("Mul", 14, "")
        op = Op(self, "Mul", schema)
        return op(*self._prepare_inputs(schema, A, B))

    T_RNN = TypeVar("T_RNN", DOUBLE, FLOAT, FLOAT16)

    T1_RNN: TypeAlias = INT32

    def RNN(
        self,
        X: T_RNN,
        W: T_RNN,
        R: T_RNN,
        B: Optional[T_RNN] = None,
        sequence_lens: Optional[T1_RNN] = None,
        initial_h: Optional[T_RNN] = None,
        *,
        activation_alpha: Optional[Sequence[float]] = None,
        activation_beta: Optional[Sequence[float]] = None,
        activations: Sequence[str] = ("Tanh", "Tanh"),
        clip: Optional[float] = None,
        direction: str = "forward",
        hidden_size: Optional[int] = None,
        layout: int = 0,
    ) -> Tuple[T_RNN, T_RNN]:
        r"""[🌐 RNN(14)](https://onnx.ai/onnx/operators/onnx__RNN.html#rnn-14 "Online Documentation")


        Computes an one-layer simple RNN. This operator is usually supported
        via some custom implementation such as CuDNN.

        Notations:

        * `X` - input tensor
        * `i` - input gate
        * `t` - time step (t-1 means previous time step)
        * `Wi` - W parameter weight matrix for input gate
        * `Ri` - R recurrence weight matrix for input gate
        * `Wbi` - W parameter bias vector for input gate
        * `Rbi` - R parameter bias vector for input gate
        * `WBi` - W parameter weight matrix for backward input gate
        * `RBi` - R recurrence weight matrix for backward input gate
        * `WBbi` - WR bias vectors for backward input gate
        * `RBbi` - RR bias vectors for backward input gate
        * `H` - Hidden state
        * `num_directions` - 2 if direction == bidirectional else 1

        Activation functions:

        * Relu(x)                - max(0, x)
        * Tanh(x)                - (1 - e^{-2x})/(1 + e^{-2x})
        * Sigmoid(x)             - 1/(1 + e^{-x})

        NOTE: Below are optional

        * Affine(x)              - alpha*x + beta
        * LeakyRelu(x)           - x if x >= 0 else alpha * x
        * ThresholdedRelu(x)     - x if x >= alpha else 0
        * ScaledTanh(x)          - alpha*Tanh(beta*x)
        * HardSigmoid(x)         - min(max(alpha*x + beta, 0), 1)
        * Elu(x)                 - x if x >= 0 else alpha*(e^x - 1)
        * Softsign(x)            - x/(1 + |x|)
        * Softplus(x)            - log(1 + e^x)

        Equations (Default: f=Tanh):

        * Ht = f(Xt*(Wi^T) + Ht-1*(Ri^T) + Wbi + Rbi)
        This operator has **optional** inputs/outputs. See `ONNX <https://github.com/onnx/onnx/blob/master/docs/IR.md>`_ for more details about the representation of optional arguments. An empty string may be used in the place of an actual argument's name to indicate a missing argument. Trailing optional arguments (those not followed by an argument that is present) may also be simply omitted.


        Args:
            X: (differentiable) The input sequences packed (and potentially padded) into
                one 3-D tensor with the shape of `[seq_length, batch_size, input_size]`.

            W: (differentiable) The weight tensor for input gate. Concatenation of `Wi`
                and `WBi` (if bidirectional). The tensor has shape `[num_directions,
                hidden_size, input_size]`.

            R: (differentiable) The recurrence weight tensor. Concatenation of `Ri` and
                `RBi` (if bidirectional). The tensor has shape `[num_directions,
                hidden_size, hidden_size]`.

            B: (optional, differentiable) The bias tensor for input gate. Concatenation
                of `[Wbi, Rbi]` and `[WBbi, RBbi]` (if bidirectional). The tensor has
                shape `[num_directions, 2*hidden_size]`. Optional: If not specified -
                assumed to be 0.

            sequence_lens: (optional, non-differentiable) Optional tensor specifying
                lengths of the sequences in a batch. If not specified - assumed all
                sequences in the batch to have length `seq_length`. It has shape
                `[batch_size]`.

            initial_h: (optional, non-differentiable) Optional initial value of the
                hidden. If not specified - assumed to be 0. It has shape
                `[num_directions, batch_size, hidden_size]`.

            activation_alpha: Optional scaling values used by some activation functions.
                The values are consumed in the order of activation functions, for
                example (f, g, h) in LSTM. Default values are the same as of
                corresponding ONNX operators.For example with LeakyRelu, the default
                alpha is 0.01.

            activation_beta: Optional scaling values used by some activation functions.
                The values are consumed in the order of activation functions, for
                example (f, g, h) in LSTM. Default values are the same as of
                corresponding ONNX operators.

            activations: One (or two if bidirectional) activation function for input
                gate. The activation function must be one of the activation functions
                specified above. Optional: Default `Tanh` if not specified.

            clip: Cell clip threshold. Clipping bounds the elements of a tensor in the
                range of [-threshold, +threshold] and is applied to the input of
                activations. No clip if not specified.

            direction: Specify if the RNN is forward, reverse, or bidirectional. Must be
                one of forward (default), reverse, or bidirectional.

            hidden_size: Number of neurons in the hidden layer

            layout: The shape format of inputs X, initial_h and outputs Y, Y_h. If 0,
                the following shapes are expected: X.shape = [seq_length, batch_size,
                input_size], Y.shape = [seq_length, num_directions, batch_size,
                hidden_size], initial_h.shape = Y_h.shape = [num_directions, batch_size,
                hidden_size]. If 1, the following shapes are expected: X.shape =
                [batch_size, seq_length, input_size], Y.shape = [batch_size, seq_length,
                num_directions, hidden_size], initial_h.shape = Y_h.shape = [batch_size,
                num_directions, hidden_size].
        """

        schema = get_schema("RNN", 14, "")
        op = Op(self, "RNN", schema)
        return op(
            *self._prepare_inputs(schema, X, W, R, B, sequence_lens, initial_h),
            activation_alpha=activation_alpha,
            activation_beta=activation_beta,
            activations=activations,
            clip=clip,
            direction=direction,
            hidden_size=hidden_size,
            layout=layout,
        )

    T_Relu = TypeVar("T_Relu", BFLOAT16, DOUBLE, FLOAT, FLOAT16, INT16, INT32, INT64, INT8)

    def Relu(self, X: T_Relu) -> T_Relu:
        r"""[🌐 Relu(14)](https://onnx.ai/onnx/operators/onnx__Relu.html#relu-14 "Online Documentation")


        Relu takes one input data (Tensor<T>) and produces one output data
        (Tensor<T>) where the rectified linear function, y = max(0, x), is applied to
        the tensor elementwise.


        Args:
            X: (differentiable) Input tensor
        """

        schema = get_schema("Relu", 14, "")
        op = Op(self, "Relu", schema)
        return op(*self._prepare_inputs(schema, X))

    T_Reshape = TypeVar(
        "T_Reshape",
        BFLOAT16,
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Reshape(self, data: T_Reshape, shape: INT64, *, allowzero: int = 0) -> T_Reshape:
        r"""[🌐 Reshape(14)](https://onnx.ai/onnx/operators/onnx__Reshape.html#reshape-14 "Online Documentation")


        Reshape the input tensor similar to numpy.reshape.
        First input is the data tensor, second input is a shape tensor which specifies the output shape. It outputs the reshaped tensor.
        At most one dimension of the new shape can be -1. In this case, the value is
        inferred from the size of the tensor and the remaining dimensions. A dimension
        could also be 0, in which case the actual dimension value is unchanged (i.e. taken
        from the input tensor). If 'allowzero' is set, and the new shape includes 0, the
        dimension will be set explicitly to zero (i.e. not taken from input tensor).
        Shape (second input) could be an empty shape, which means converting to a scalar.
        The input tensor's shape and the output tensor's shape are required to have the same number of elements.

        If the attribute 'allowzero' is set, it is invalid for the specified shape to
        contain both a zero value and -1, as the value of the dimension corresponding
        to -1 cannot be determined uniquely.


        Args:
            data: (differentiable) An input tensor.

            shape: (non-differentiable) Specified shape for output.

            allowzero: (Optional) By default, when any value in the 'shape' input is
                equal to zero the corresponding dimension value is copied from the input
                tensor dynamically. allowzero=1 indicates that if any value in the
                'shape' input is set to zero, the zero value is honored, similar to
                NumPy.
        """

        schema = get_schema("Reshape", 14, "")
        op = Op(self, "Reshape", schema)
        return op(*self._prepare_inputs(schema, data, shape), allowzero=allowzero)

    T_Sub = TypeVar(
        "T_Sub",
        BFLOAT16,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Sub(self, A: T_Sub, B: T_Sub) -> T_Sub:
        r"""[🌐 Sub(14)](https://onnx.ai/onnx/operators/onnx__Sub.html#sub-14 "Online Documentation")


        Performs element-wise binary subtraction (with Numpy-style broadcasting support).

        This operator supports **multidirectional (i.e., Numpy-style) broadcasting**; for more details please check `Broadcasting in ONNX <https://github.com/onnx/onnx/blob/master/docs/Broadcasting.md>`_.

        (Opset 14 change): Extend supported types to include uint8, int8, uint16, and int16.


        Args:
            A: (differentiable) First operand.

            B: (differentiable) Second operand.
        """

        schema = get_schema("Sub", 14, "")
        op = Op(self, "Sub", schema)
        return op(*self._prepare_inputs(schema, A, B))

    T_Trilu = TypeVar(
        "T_Trilu",
        BFLOAT16,
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Trilu(self, input: T_Trilu, k: Optional[INT64] = None, *, upper: int = 1) -> T_Trilu:
        r"""[🌐 Trilu(14)](https://onnx.ai/onnx/operators/onnx__Trilu.html#trilu-14 "Online Documentation")


        Given a 2-D matrix or batches of 2-D matrices, returns the upper or lower triangular part of the tensor(s).
        The attribute "upper" determines whether the upper or lower part is retained. If set to true,
        the upper triangular matrix is retained. Lower triangular matrix is retained otherwise.
        Default value for the "upper" attribute is true.
        Trilu takes one input tensor of shape [*, N, M], where * is zero or more batch dimensions. The upper triangular part consists
        of the elements on and above the given diagonal (k). The lower triangular part consists of elements on and below the diagonal.
        All other elements in the matrix are set to zero.
        If k = 0, the triangular part on and above/below the main diagonal is retained.
        If upper is set to true, a positive k retains the upper triangular matrix excluding the main diagonal and (k-1) diagonals above it.
        A negative k value retains the main diagonal and |k| diagonals below it.
        If upper is set to false, a positive k retains the lower triangular matrix including the main diagonal and k diagonals above it.
        A negative k value excludes the main diagonal and (|k|-1) diagonals below it.


        Args:
            input: (differentiable) Input tensor of rank 2 or higher.

            k: (optional, non-differentiable) A 0-D tensor containing a single value
                corresponding to the number diagonals above or below the main diagonal
                to exclude or include. Default value is 0 if it's not specified.

            upper: Boolean. Indicates whether upper or lower part of matrix is retained.
                Default is true.
        """

        schema = get_schema("Trilu", 14, "")
        op = Op(self, "Trilu", schema)
        return op(*self._prepare_inputs(schema, input, k), upper=upper)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import unittest

import numpy as np
import onnx.checker
import onnx.shape_inference
import onnxruntime

from onnxscript import ir
from onnxscript.rewriter.onnxruntime.bfloat16_utils import bfloat16_converter


class Bfloat16ConversionTest(unittest.TestCase):
    def setUp(self) -> None:
        self.v0 = ir.val(name="v0", shape=ir.Shape([2, 3, 4]))
        self.v0.dtype = ir.DataType.BFLOAT16
        self.v1 = ir.val(name="v1", shape=ir.Shape([2, 3, 4]))
        self.v1.dtype = ir.DataType.BFLOAT16
        self.v2 = ir.val(name="v2", shape=ir.Shape([2, 3, 4]))
        self.v2.dtype = ir.DataType.BFLOAT16

        self.add_node = ir.Node("", "Add", inputs=(self.v0, self.v1), num_outputs=1)
        self.add_node.outputs[0].dtype = ir.DataType.BFLOAT16
        self.mul_node = ir.Node(
            "", "Mul", inputs=(self.add_node.outputs[0], self.v2), num_outputs=1
        )
        self.mul_node.outputs[0].dtype = ir.DataType.BFLOAT16
        self.graph = ir.Graph(
            name="bfloat16_conversion_test",
            inputs=(self.v0, self.v1, self.v2),
            outputs=(self.add_node.outputs[0], self.mul_node.outputs[0]),
            nodes=(self.add_node, self.mul_node),
            opset_imports={"": 18},
        )
        self.original_output_names = [output.name for output in self.graph.outputs]
        self.model = ir.Model(
            graph=self.graph,
            ir_version=8,
            producer_name="bfloat16_conversion_test",
        )
        bfloat16_converter.dtype_adapter_for_bfloat16_model(self.model)

    def test_input_and_output_are_float16(self):
        for input in self.model.graph.inputs:
            self.assertEqual(input.dtype, ir.DataType.FLOAT16)
        for output in self.model.graph.outputs:
            self.assertEqual(output.dtype, ir.DataType.FLOAT16)

    def test_cast_nodes_are_inserted(self):
        cast_node_count = 0
        for node in self.model.graph:
            if node.op_type == "Cast":
                cast_node_count += 1
        self.assertEqual(cast_node_count, 5)

        for input in self.model.graph.inputs:
            for input_user, _ in input.uses():
                self.assertEqual(input_user.op_type, "Cast")
                self.assertEqual(input_user.outputs[0].dtype, ir.DataType.BFLOAT16)
        for output in self.model.graph.outputs:
            self.assertEqual(output.producer().op_type, "Cast")
            self.assertEqual(output.producer().inputs[0].dtype, ir.DataType.BFLOAT16)

    def test_graph_output_name_is_preserved(self):
        self.assertEqual(
            [output.name for output in self.model.graph.outputs],
            self.original_output_names,
        )

    def test_bfloat16_converted_model_runtime(self):
        model_proto = ir.serde.serialize_model(self.model)
        model_proto_filled_shape_type = onnx.shape_inference.infer_shapes(
            model_proto, check_type=True, strict_mode=True, data_prop=True
        )
        onnx.checker.check_model(model_proto_filled_shape_type, full_check=True)
        try:
            ort_session = onnxruntime.InferenceSession(
                model_proto_filled_shape_type.SerializeToString(),
                providers=["CPUExecutionProvider"],
            )
            v0 = np.random.randn(2, 3, 4).astype(np.float16)
            v1 = np.random.randn(2, 3, 4).astype(np.float16)
            v2 = np.random.randn(2, 3, 4).astype(np.float16)
            ort_inputs = {"v0": v0, "v1": v1, "v2": v2}
            ort_outputs = ort_session.run(None, ort_inputs)
            expected_output = (v0 + v1) * v2
            np.testing.assert_allclose(ort_outputs[0], expected_output, rtol=1e-2, atol=1e-2)
            np.testing.assert_allclose(ort_outputs[1], expected_output, rtol=1e-2, atol=1e-2)
        except Exception as e:
            self.assertIn(
                "[ONNXRuntimeError] : 9 : NOT_IMPLEMENTED : Could not find an implementation for Add(14)",
                str(e),
            )


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import unittest

import onnx.parser

from onnxscript import ir
from onnxscript.rewriter.rules.common import _gemm_to_matmul_add


class ReshapeGemmReshapeTest(unittest.TestCase):
    def test_reshape_gemm_reshape_replace_when_nd_inputs_are_broadcastable(self):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[1, 4, 512, 512] input_x, float[4, 512, 64] input_y, float[4, 512, 64] input_z) => (float[1, 4, 512, 64] output)
            {
                shape_a = Constant<value: tensor = int64[3] {4, 512, 512}>()
                reshape_x = Reshape (input_x, shape_a)
                gemm = Gemm<alpha=1.0, beta=1.0> (reshape_x, input_y, input_z)
                shape_d = Constant<value: tensor = int64[4] {1, 4, 512, 64}>()
                output = Reshape (gemm, shape_d)
            }
        """
        )

        model = ir.serde.deserialize_model(model_proto)
        count = _gemm_to_matmul_add.gemm_to_matmul_add_rule.apply_to_model(model)
        self.assertEqual(count, 1)
        self.assertEqual(len(model.graph), 4)

    def test_reshape_gemm_reshape_replace_when_nd_inputs_are_broadcastable_in_nested_function(
        self,
    ):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17, "pkg.custom": 1]>
            agraph (float[1, 4, 512, 512] input_x, float[4, 512, 64] input_y, float[4, 512, 64] input_z) => (float[1, 4, 512, 64] output)
            {
                output = afunction (input_x, input_y, input_z)
            }
            <domain: "pkg.custom", opset_import: [ "" : 17]>
            afunction (input_x, input_y, input_z) => (output)
            {
                shape_a = Constant<value: tensor = int64[3] {4, 512, 512}>()
                reshape_x = Reshape (input_x, shape_a)
                gemm = Gemm<alpha=1.0, beta=1.0> (reshape_x, input_y, input_z)
                shape_d = Constant<value: tensor = int64[4] {1, 4, 512, 64}>()
                output = Reshape (gemm, shape_d)
            }
        """
        )
        # Hack to put value_info in since parser does not support this experimental naming format
        model_proto.graph.value_info.append(
            onnx.helper.make_tensor_value_info(
                "pkg.custom::afunction/input_x",
                onnx.TensorProto.FLOAT,
                [1, 4, 512, 512],
            )
        )
        model_proto.graph.value_info.append(
            onnx.helper.make_tensor_value_info(
                "pkg.custom::afunction/input_y", onnx.TensorProto.FLOAT, [4, 512, 64]
            )
        )
        model_proto.graph.value_info.append(
            onnx.helper.make_tensor_value_info(
                "pkg.custom::afunction/input_z", onnx.TensorProto.FLOAT, [1, 4, 512, 64]
            )
        )

        model = ir.serde.deserialize_model(model_proto)
        count = _gemm_to_matmul_add.gemm_to_matmul_add_rule.apply_to_model(model)
        self.assertEqual(count, 1)
        self.assertEqual(len(model.functions), 1)
        self.assertEqual(len(model.functions[("pkg.custom", "afunction", "")]), 4)
        self.assertEqual(model.functions[("pkg.custom", "afunction", "")][2].op_type, "MatMul")
        self.assertEqual(model.functions[("pkg.custom", "afunction", "")][3].op_type, "Add")

    def test_reshape_gemm_reshape_remain_when_input_last_dim_and_second_last_dim_not_matched(
        self,
    ):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[1, 4, 512, 512] input_x, float[4, 256, 64] input_y, float[4, 512, 64] input_z) => (float[1, 4, 512, 64] output)
            {
                shape_a = Constant<value: tensor = int64[3] {4, 512, 512}>()
                reshape_x = Reshape (input_x, shape_a)
                gemm = Gemm<alpha=1.0, beta=1.0> (reshape_x, input_y, input_z)
                shape_d = Constant<value: tensor = int64[4] {1, 4, 512, 64}>()
                output = Reshape (gemm, shape_d)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _gemm_to_matmul_add.gemm_to_matmul_add_rule.apply_to_model(model)
        self.assertEqual(count, 0)
        self.assertEqual(len(model.graph), 5)

    def test_reshape_gemm_reshape_remain_when_inputs_are_not_broadcastable(
        self,
    ):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[2, 2, 512, 512] input_x, float[4, 512, 64] input_y, float[4, 512, 64] input_z) => (float[1, 4, 512, 64] output)
            {
                shape_a = Constant<value: tensor = int64[3] {4, 512, 512}>()
                reshape_x = Reshape (input_x, shape_a)
                gemm = Gemm<alpha=1.0, beta=1.0> (reshape_x, input_y, input_z)
                shape_d = Constant<value: tensor = int64[4] {1, 4, 512, 64}>()
                output = Reshape (gemm, shape_d)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _gemm_to_matmul_add.gemm_to_matmul_add_rule.apply_to_model(model)
        self.assertEqual(count, 0)
        self.assertEqual(len(model.graph), 5)

    def test_reshape_gemm_reshape_replace_when_inputs_are_broadcastable_with_one_in_dims(
        self,
    ):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[4, 512, 512] input_x, float[1, 4, 512, 64] input_y, float[1, 4, 512, 64] input_z) => (float[1, 4, 512, 64] output)
            {
                shape_a = Constant<value: tensor = int64[3] {4, 512, 512}>()
                reshape_x = Reshape (input_x, shape_a)
                gemm = Gemm<alpha=1.0, beta=1.0> (reshape_x, input_y, input_z)
                shape_d = Constant<value: tensor = int64[4] {1, 4, 512, 64}>()
                output = Reshape (gemm, shape_d)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _gemm_to_matmul_add.gemm_to_matmul_add_rule.apply_to_model(model)
        self.assertEqual(count, 1)
        self.assertEqual(len(model.graph), 4)
        self.assertEqual(model.graph[2].op_type, "MatMul")
        self.assertEqual(model.graph[3].op_type, "Add")

    def test_reshape_gemm_reshape_replace_when_first_input_is_one_dimension_and_broadcastable(
        self,
    ):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[4] input_x, float[2, 3, 4, 5] input_y, float[2, 3, 5] input_z) => (float[2, 3, 5] output)
            {
                shape_a = Constant<value: tensor = int64[2] {1, 4}>()
                reshape_x = Reshape (input_x, shape_a)
                gemm = Gemm<alpha=1.0, beta=1.0> (reshape_x, input_y, input_z)
                shape_d = Constant<value: tensor = int64[3] {2, 3, 5}>()
                output = Reshape (gemm, shape_d)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _gemm_to_matmul_add.gemm_to_matmul_add_rule.apply_to_model(model)
        self.assertEqual(count, 1)
        self.assertEqual(len(model.graph), 4)
        self.assertEqual(model.graph[2].op_type, "MatMul")
        self.assertEqual(model.graph[3].op_type, "Add")

    def test_reshape_gemm_reshape_remain_when_first_input_is_one_dimension_and_not_broadcastable(
        self,
    ):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[8] input_x, float[2, 3, 4, 5] input_y, float[2, 3, 5] input_z) => (float[2, 3, 5] output)
            {
                shape_a = Constant<value: tensor = int64[2] {2, 4}>()
                reshape_x = Reshape (input_x, shape_a)
                gemm = Gemm<alpha=1.0, beta=1.0> (reshape_x, input_y, input_z)
                shape_d = Constant<value: tensor = int64[3] {2, 3, 5}>()
                output = Reshape (gemm, shape_d)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _gemm_to_matmul_add.gemm_to_matmul_add_rule.apply_to_model(model)
        self.assertEqual(count, 0)
        self.assertEqual(len(model.graph), 5)

    def test_reshape_gemm_reshape_replace_when_second_input_is_one_dimension_and_broadcastable(
        self,
    ):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[2, 3, 5, 4] input_x, float[4] input_y, float[2, 3, 5] input_z) => (float[2, 3, 5] output)
            {
                shape_a = Constant<value: tensor = int64[4] {2, 3, 5, 4}>()
                reshape_x = Reshape (input_x, shape_a)
                gemm = Gemm<alpha=1.0, beta=1.0> (reshape_x, input_y, input_z)
                shape_d = Constant<value: tensor = int64[3] {2, 3, 5}>()
                output = Reshape (gemm, shape_d)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _gemm_to_matmul_add.gemm_to_matmul_add_rule.apply_to_model(model)
        self.assertEqual(count, 1)
        self.assertEqual(len(model.graph), 4)
        self.assertEqual(model.graph[2].op_type, "MatMul")
        self.assertEqual(model.graph[3].op_type, "Add")

    def test_reshape_gemm_reshape_remain_when_second_input_is_one_dimension_and_not_broadcastable(
        self,
    ):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[2, 3, 5, 4] input_x, float[10] input_y, float[2, 3, 5] input_z) => (float[2, 3, 5] output)
            {
                shape_a = Constant<value: tensor = int64[4] {2, 3, 5, 4}>()
                reshape_x = Reshape (input_x, shape_a)
                gemm = Gemm<alpha=1.0, beta=1.0> (reshape_x, input_y, input_z)
                shape_d = Constant<value: tensor = int64[3] {2, 3, 5}>()
                output = Reshape (gemm, shape_d)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _gemm_to_matmul_add.gemm_to_matmul_add_rule.apply_to_model(model)
        self.assertEqual(count, 0)
        self.assertEqual(len(model.graph), 5)

    def test_reshape_gemm_reshape_replaces_when_inputs_are_two_dimensional_and_broadcastable(
        self,
    ):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[3, 5] input_x, float[5, 10] input_y, float[3, 10] input_z) => (float[3, 10] output)
            {
                shape_a = Constant<value: tensor = int64[2] {3, 5}>()
                reshape_x = Reshape (input_x, shape_a)
                gemm = Gemm<alpha=1.0, beta=1.0> (reshape_x, input_y, input_z)
                shape_d = Constant<value: tensor = int64[2] {3, 10}>()
                output = Reshape (gemm, shape_d)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        replacement_count = _gemm_to_matmul_add.gemm_to_matmul_add_rule.apply_to_model(model)
        self.assertEqual(replacement_count, 1)
        self.assertEqual(len(model.graph), 4)

    def test_reshape_gemm_reshape_remain_when_inputs_are_two_dimension_and_not_broadcastable(
        self,
    ):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[5, 3] input_x, float[5, 10] input_y, float[3, 10] input_z) => (float[3, 10] output)
            {
                shape_a = Constant<value: tensor = int64[2] {3, 5}>()
                reshape_x = Reshape (input_x, shape_a)
                gemm = Gemm<alpha=1.0, beta=1.0> (reshape_x, input_y, input_z)
                shape_d = Constant<value: tensor = int64[2] {3, 10}>()
                output = Reshape (gemm, shape_d)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _gemm_to_matmul_add.gemm_to_matmul_add_rule.apply_to_model(model)
        self.assertEqual(count, 0)
        self.assertEqual(len(model.graph), 5)

    def test_reshape_gemm_reshape_remain_when_output_is_not_matmul_broadcasted(
        self,
    ):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[2, 3, 5, 4] input_x, float[5] input_y, float[2, 3, 5] input_z) => (float[2, 4, 6] output)
            {
                shape_a = Constant<value: tensor = int64[4] {2, 3, 5, 4}>()
                reshape_x = Reshape (input_x, shape_a)
                gemm = Gemm<alpha=1.0, beta=1.0> (reshape_x, input_y, input_z)
                shape_d = Constant<value: tensor = int64[3] {2, 4, 6}>()
                output = Reshape (gemm, shape_d)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _gemm_to_matmul_add.gemm_to_matmul_add_rule.apply_to_model(model)
        self.assertEqual(count, 0)
        self.assertEqual(len(model.graph), 5)


if __name__ == "__main__":
    unittest.main()



# --------------------------------------------------------------------------
# ⚠️ WARNING - AUTO-GENERATED CODE - DO NOT EDIT ⚠️
# ⚙️ Generated by 'python -m opgen'
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# pylint: disable=W0221,W0222,R0901,W0237
# mypy: disable-error-code=override
# ruff: noqa: N801
# --------------------------------------------------------------------------

from __future__ import annotations

from typing import Optional, Sequence, TypeVar

from onnx import TensorProto
from onnx.defs import get_schema

from onnxscript.onnx_opset._impl.opset_ai_onnx_ml4 import Opset_ai_onnx_ml4
from onnxscript.onnx_types import DOUBLE, FLOAT, FLOAT16
from onnxscript.values import Op, Opset


class Opset_ai_onnx_ml5(Opset_ai_onnx_ml4):
    def __new__(cls):
        return Opset.__new__(cls, "ai.onnx.ml", 5)

    T_TreeEnsemble = TypeVar("T_TreeEnsemble", DOUBLE, FLOAT, FLOAT16)

    def TreeEnsemble(
        self,
        X: T_TreeEnsemble,
        *,
        aggregate_function: int = 1,
        leaf_targetids: Sequence[int],
        leaf_weights: TensorProto,
        membership_values: Optional[TensorProto] = None,
        n_targets: Optional[int] = None,
        nodes_falseleafs: Sequence[int],
        nodes_falsenodeids: Sequence[int],
        nodes_featureids: Sequence[int],
        nodes_hitrates: Optional[TensorProto] = None,
        nodes_missing_value_tracks_true: Optional[Sequence[int]] = None,
        nodes_modes: TensorProto,
        nodes_splits: TensorProto,
        nodes_trueleafs: Sequence[int],
        nodes_truenodeids: Sequence[int],
        post_transform: int = 0,
        tree_roots: Sequence[int],
    ) -> T_TreeEnsemble:
        r"""[🌐 ai.onnx.ml::TreeEnsemble(5)](https://onnx.ai/onnx/operators/onnx_aionnxml_TreeEnsemble.html#treeensemble-5 "Online Documentation")


            Tree Ensemble operator.  Returns the regressed values for each input in a batch.
            Inputs have dimensions `[N, F]` where `N` is the input batch size and `F` is the number of input features.
            Outputs have dimensions `[N, num_targets]` where `N` is the batch size and `num_targets` is the number of targets, which is a configurable attribute.

            The encoding of this attribute is split along interior nodes and the leaves of the trees. Notably, attributes with the prefix `nodes_*` are associated with interior nodes, and attributes with the prefix `leaf_*` are associated with leaves.
            The attributes `nodes_*` must all have the same length and encode a sequence of tuples, as defined by taking all the `nodes_*` fields at a given position.

            All fields prefixed with `leaf_*` represent tree leaves, and similarly define tuples of leaves and must have identical length.

            This operator can be used to implement both the previous `TreeEnsembleRegressor` and `TreeEnsembleClassifier` nodes.
            The `TreeEnsembleRegressor` node maps directly to this node and requires changing how the nodes are represented.
            The `TreeEnsembleClassifier` node can be implemented by adding a `ArgMax` node after this node to determine the top class.
            To encode class labels, a `LabelEncoder` or `GatherND` operator may be used.


        Args:
            X: Input of shape [Batch Size, Number of Features]

            aggregate_function: Defines how to aggregate leaf values within a target.
                <br>One of 'AVERAGE' (0) 'SUM' (1) 'MIN' (2) 'MAX (3) defaults to 'SUM'
                (1)

            leaf_targetids: The index of the target that this leaf contributes to (this
                must be in range `[0, n_targets)`).

            leaf_weights: The weight for each leaf.

            membership_values: Members to test membership of for each set membership
                node. List all of the members to test again in the order that the
                'BRANCH_MEMBER' mode appears in `node_modes`, delimited by `NaN`s. Will
                have the same number of sets of values as nodes with mode
                'BRANCH_MEMBER'. This may be omitted if the node doesn't contain any
                'BRANCH_MEMBER' nodes.

            n_targets: The total number of targets.

            nodes_falseleafs: 1 if false branch is leaf for each node and 0 if an
                interior node. To represent a tree that is a leaf (only has one node),
                one can do so by having a single `nodes_*` entry with true and false
                branches referencing the same `leaf_*` entry

            nodes_falsenodeids: If `nodes_falseleafs` is false at an entry, this
                represents the position of the false branch node. This position can be
                used to index into a `nodes_*` entry. If `nodes_falseleafs` is false, it
                is an index into the leaf_* attributes.

            nodes_featureids: Feature id for each node.

            nodes_hitrates: Popularity of each node, used for performance and may be
                omitted.

            nodes_missing_value_tracks_true: For each node, define whether to follow the
                true branch (if attribute value is 1) or false branch (if attribute
                value is 0) in the presence of a NaN input feature. This attribute may
                be left undefined and the default value is false (0) for all nodes.

            nodes_modes: The comparison operation performed by the node. This is encoded
                as an enumeration of 0 ('BRANCH_LEQ'), 1 ('BRANCH_LT'), 2
                ('BRANCH_GTE'), 3 ('BRANCH_GT'), 4 ('BRANCH_EQ'), 5 ('BRANCH_NEQ'), and
                6 ('BRANCH_MEMBER'). Note this is a tensor of type uint8.

            nodes_splits: Thresholds to do the splitting on for each node with mode that
                is not 'BRANCH_MEMBER'.

            nodes_trueleafs: 1 if true branch is leaf for each node and 0 an interior
                node. To represent a tree that is a leaf (only has one node), one can do
                so by having a single `nodes_*` entry with true and false branches
                referencing the same `leaf_*` entry

            nodes_truenodeids: If `nodes_trueleafs` is false at an entry, this
                represents the position of the true branch node. This position can be
                used to index into a `nodes_*` entry. If `nodes_trueleafs` is false, it
                is an index into the leaf_* attributes.

            post_transform: Indicates the transform to apply to the score. <br>One of
                'NONE' (0), 'SOFTMAX' (1), 'LOGISTIC' (2), 'SOFTMAX_ZERO' (3) or
                'PROBIT' (4), defaults to 'NONE' (0)

            tree_roots: Index into `nodes_*` for the root of each tree. The tree
                structure is derived from the branching of each node.
        """

        schema = get_schema("TreeEnsemble", 5, "ai.onnx.ml")
        op = Op(self, "TreeEnsemble", schema)
        return op(
            *self._prepare_inputs(schema, X),
            aggregate_function=aggregate_function,
            leaf_targetids=leaf_targetids,
            leaf_weights=leaf_weights,
            membership_values=membership_values,
            n_targets=n_targets,
            nodes_falseleafs=nodes_falseleafs,
            nodes_falsenodeids=nodes_falsenodeids,
            nodes_featureids=nodes_featureids,
            nodes_hitrates=nodes_hitrates,
            nodes_missing_value_tracks_true=nodes_missing_value_tracks_true,
            nodes_modes=nodes_modes,
            nodes_splits=nodes_splits,
            nodes_trueleafs=nodes_trueleafs,
            nodes_truenodeids=nodes_truenodeids,
            post_transform=post_transform,
            tree_roots=tree_roots,
        )



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""PyTorch-like module interface for building ONNX graphs."""

from onnxscript.nn._module import Module
from onnxscript.nn._module_list import ModuleList
from onnxscript.nn._parameter import Parameter
from onnxscript.nn._sequential import Sequential

__all__ = ["Module", "ModuleList", "Parameter", "Sequential"]



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from onnxscript.rewriter._rewrite_rule import RewriteRule, RewriteRuleSet

# TODO: Support 1-D constant tensors
# https://github.com/microsoft/onnx-rewriter/issues/186


# Pattern to match against
def mul_by_1(op, x):
    return x * 1


def add_0(op, x):
    return x + 0


def sub_0(op, x):
    return x - 0


def div_by_1(op, x):
    return x / 1


def dropout_zero(op, x):
    return op.Dropout(x, ratio=0.0)


def dropout_inference(op, x):
    return op.Dropout(x, training_mode=False)


# Replacement
def identity(op, x, **_):
    return op.Identity(x)


mul_by_1_rule = RewriteRule(mul_by_1, identity)
add_0_rule = RewriteRule(add_0, identity)
sub_0_rule = RewriteRule(sub_0, identity)
div_by_1_rule = RewriteRule(div_by_1, identity)
dropout_zero_rule = RewriteRule(dropout_zero, identity)
dropout_inference_rule = RewriteRule(dropout_inference, identity)
# TODO: Include Mul by 0, 0 by Mul, 0 by Div? Those would be 0s, but not no-ops

rules = RewriteRuleSet(
    [
        *mul_by_1_rule.commute(),
        *add_0_rule.commute(),
        sub_0_rule,
        div_by_1_rule,
        dropout_zero_rule,
        dropout_inference_rule,
    ]
)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import inspect
import typing
import unittest

import onnxscript
from onnxscript._internal import values


class TracedOnnxFunctionTest(unittest.TestCase):
    def test_init(self):
        def function(input1, input2, attr1: int, attr2: int = 1):
            return input1 + input2 + attr1 + attr2

        opset = values.Opset("test", 1)
        traced_function = values.TracedOnnxFunction(opset, function)
        self.assertEqual(traced_function.opset, opset)
        self.assertEqual(traced_function.name, function.__name__)
        self.assertEqual(traced_function.func, function)

    def test_it_preserves_the_function_signature(self):
        opset = values.Opset("test", 1)

        def function(input1, input2, attr1: int, attr2: float, input3, attr3: str = "default"):
            return opset.CustomOp(input1 + input2, input3, attr1, attr2, attr3)

        traced_function = values.TracedOnnxFunction(opset, function)
        signature = inspect.signature(traced_function)
        self.assertEqual(signature.parameters["input1"].name, "input1")
        self.assertEqual(signature.parameters["input2"].name, "input2")
        self.assertEqual(signature.parameters["attr1"].name, "attr1")
        self.assertEqual(signature.parameters["attr2"].name, "attr2")
        self.assertEqual(signature.parameters["input3"].name, "input3")
        self.assertEqual(signature.parameters["attr3"].name, "attr3")

        annotations = typing.get_type_hints(traced_function)
        self.assertEqual(annotations["attr1"], int)
        self.assertEqual(annotations["attr2"], float)
        self.assertEqual(annotations["attr3"], str)


class OnnxFunctionTest(unittest.TestCase):
    def test_it_preserves_the_function_signature(self):
        opset = values.Opset("test", 1)

        @onnxscript.script(default_opset=opset)
        def function(input1, input2, attr1: int, attr2: float, input3, attr3: str = "default"):
            return opset.CustomOp(input1 + input2, input3, attr1, attr2, attr3)

        signature = inspect.signature(function)
        self.assertEqual(signature.parameters["input1"].name, "input1")
        self.assertEqual(signature.parameters["input2"].name, "input2")
        self.assertEqual(signature.parameters["attr1"].name, "attr1")
        self.assertEqual(signature.parameters["attr2"].name, "attr2")
        self.assertEqual(signature.parameters["input3"].name, "input3")
        self.assertEqual(signature.parameters["attr3"].name, "attr3")

        annotations = typing.get_type_hints(function)
        self.assertEqual(annotations["attr1"], int)
        self.assertEqual(annotations["attr2"], float)
        self.assertEqual(annotations["attr3"], str)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

from onnxscript.ir import _tape
from onnxscript.rewriter._basics import MatchingTracer, MatchResult, MatchStatus
from onnxscript.rewriter._matcher import PatternMatcher, SimplePatternMatcher
from onnxscript.rewriter._pattern_ir import (
    ANY_VALUE,
    AttrVar,
    Constant,
    OpsetPatternBuilder,
    OrValue,
    Var,
    pattern_builder,
    torch_module_op,
)
from onnxscript.rewriter._rewrite_rule import (
    Pattern,
    PatternBase,
    RewriteRule,
    RewriteRuleClassBase,
    RewriteRuleSet,
)

RewriterContext = _tape.Builder

__all__ = [
    "ANY_VALUE",
    "AttrVar",
    "OrValue",
    "Constant",
    "OpsetPatternBuilder",
    "pattern_builder",
    "PatternBase",
    "Pattern",
    "RewriteRule",
    "RewriteRuleClassBase",
    "RewriteRuleSet",
    "RewriterContext",
    "MatchingTracer",
    "MatchResult",
    "MatchStatus",
    "PatternMatcher",
    "SimplePatternMatcher",
    "torch_module_op",
    "Var",
]



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Experimental flags.

NOTE: These flags are experimental only. Any flag here can be removed at any
time without notice.
"""

import logging
import os

logger = logging.getLogger(__name__)


def _load_boolean_flag(
    name: str,
    *,
    this_will: str,
    deprecated: bool = False,
    default: bool = False,
) -> bool:
    """Load a boolean flag from environment variable.

    Args:
        name: The name of the environment variable.
        this_will: A string that describes what this flag will do.
        deprecated: Whether this flag is deprecated.
        default: The default value if envvar not defined.
    """
    undefined = os.getenv(name) is None
    state = os.getenv(name) == "1"
    if state:
        if deprecated:
            logger.error(
                "Experimental flag %s is deprecated. Please remove it from your environment.",
                name,
            )
        else:
            logger.warning("Experimental flag %s is enabled. This will %s.", name, this_will)
    if undefined:
        state = default
    return state


EXPERIMENTAL_INITIALIZERS_AS_INPUTS: bool = _load_boolean_flag(
    "TORCHLIB_EXPERIMENTAL_INITIALIZERS_AS_INPUTS",
    this_will="make initializers as inputs to the model graph",
)
EXPERIMENTAL_PREFER_TRACING: bool = _load_boolean_flag(
    "TORCHLIB_EXPERIMENTAL_PREFER_TRACING",
    this_will="trace all traceable functions to fold if branches and collapse constant expressions",
    default=True,
)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import onnx_ir as ir

from onnxscript.rewriter import _fusion_utils, _ir_utils, pattern

"""
Layer Normalization fusion optimization.

This module contains rewrite rules for fusing Layer Normalization patterns into the
ONNX LayerNormalization operator.

Layer Normalization performs normalization over the last D dimensions as specified by the axis.
The computation follows: Y = scale * (X - mean) / sqrt(variance + epsilon) + bias

Key points for the fusion optimization:
* Following restrictions from opset 17 LayerNormalization:
* Input, scale, and bias must be of same type T in {float16, bfloat16, float, double}
* The normalization can be done in a different precision than the input type (bfloat16 or float),
which is also the precision of the output mean/invstddev
"""

# input types permitted by LayerNormalization op (ONNX Opset 17)
LAYER_NORM_INPUT_TYPES = frozenset(
    [
        ir.DataType.FLOAT,
        ir.DataType.FLOAT16,
        ir.DataType.BFLOAT16,
        ir.DataType.DOUBLE,
    ]
)

# Compute types permitted by LayerNormalization op (ONNX Opset 17), aka stash_type.
LAYER_NORM_COMPUTE_TYPES = frozenset([ir.DataType.FLOAT, ir.DataType.DOUBLE])


class LayerNormFusion(pattern.RewriteRuleClassBase):
    """Fuse LayerNorm pattern into LayerNormalization op."""

    def pattern(self, op, x, scale, epsilon):
        # Compute mean: Mean = ReduceMean(X, axes=normalized_axes)
        # TODO: support axes attribute too
        mean = op.ReduceMean(x, [-1], keepdims=1)

        # Compute deviation: D = Sub(X, Mean)
        deviation = op.Sub(x, mean)

        # Compute squared deviation: DD = Mul(D, D)
        deviation_squared = pattern.OrValue(
            [
                op.Mul(deviation, deviation),
                op.Pow(deviation, 2),
            ]
        )

        # Compute variance: Var = ReduceMean(DD, axes=normalized_axes)
        variance = op.ReduceMean(deviation_squared, [-1], keepdims=1)

        # Add epsilon: VarEps = Add(Var, epsilon)
        variance_plus_epsilon = op.Add(variance, epsilon)

        # Compute standard deviation: StdDev = Sqrt(VarEps)
        std_dev = op.Sqrt(variance_plus_epsilon)

        # Compute reciprocal: InvStdDev = Reciprocal(StdDev)
        # Normalize: Normalized = Mul(D, InvStdDev)

        inv_std_dev = op.Reciprocal(std_dev)
        normalized = pattern.OrValue(
            [op.Mul(deviation, inv_std_dev), op.Div(deviation, std_dev)]
        )

        # Scale: NormalizedScaled = Mul(Normalized, Scale)
        normalized_scaled = op.Mul(normalized, scale)

        return normalized_scaled

    def check(self, context, x, epsilon, **_) -> pattern.MatchResult:  # type: ignore[name-defined]
        """Check if the pattern matches conditions for use of LayerNormalization op."""
        check_result = pattern.MatchResult()

        # Type validation:
        if x.dtype not in LAYER_NORM_COMPUTE_TYPES:
            return check_result.fail("Input is not a float type.", x)
        self._stash_type = x.dtype

        # Check that epsilon is a scalar constant
        epsilon_value = _ir_utils.get_singleton_value(epsilon)
        if epsilon_value is None:
            return check_result.fail("Epsilon is not a constant scalar.", epsilon)
        # Epsilon is guaranteed to be same type as x (float or double, in this pattern)
        self._epsilon = float(epsilon_value)

        return check_result

    def rewrite(self, op, x, scale, epsilon, **_):
        return op.LayerNormalization(
            x,
            scale,
            axis=-1,
            epsilon=self._epsilon,
            stash_type=self._stash_type,
        )


class LayerNormBiasFusion(pattern.RewriteRuleClassBase):
    """Fuse LayerNorm => Add into LayerNorm with bias."""

    def pattern(self, op, x, scale, bias):
        return op.LayerNormalization(x, scale, _outputs=["normalized"]) + bias

    def rewrite(self, op, x, scale, bias, normalized):
        layernorm_node = normalized.producer()
        attributes = layernorm_node.attributes
        num_outputs = len(layernorm_node.outputs)
        return op.LayerNormalization(x, scale, bias, _outputs=num_outputs, **attributes)


# Create rules for both with and without bias
_layer_norm_rule = LayerNormFusion.rule()
_layer_norm_with_bias_rule = LayerNormBiasFusion.rule()

layer_normalization_rules = [_layer_norm_rule, _layer_norm_with_bias_rule]
layer_normalization_ruleset = pattern.RewriteRuleSet(layer_normalization_rules)

fuse_layer_normalization = _fusion_utils.apply_fusion_rules(layer_normalization_ruleset)
